unit GsmHelper;

interface

uses SysUtils, Math, Utilities;

const
  InitialTlvArraySize = 19;

type
  TBandType     = (btUnknown,btGsm900,btEGsm,btGsm1800,btUmts,btGsmR);

  TValueRange   = set of Byte;

  TTlvObject  = record
                Tag   : word;
                Value : string;
                Tlv   : string;
                end;

  TTlvObjects = array of TTlvObject;

  TTlvContainer = class(TObject)
  private
    FTlvArray : TTlvObjects;
    FNextFree : integer;
    function GetTlvObject(i: integer): TTlvObject;
  public
    property  TlvArray[i: integer]: TTlvObject read GetTlvObject; default;

    procedure Add(const New: TTlvObject); overload;
    procedure Add(const New: string);     overload;
    procedure Insert(const InsertAt: integer; const From: TTlvContainer);
    function  AllObjects: string;
    function  Count: integer;
    procedure Clear;

    constructor Create;
    destructor  Destroy; override;
  end;

  function FilterNumbers(const From: string): string;

{ conversions which are closely related to GSM/3G }

      {extracts the GSM bitstuffed 7bit data to 8bit data}
  function SevenToEightBit(const Input: string): string;

      {does the GSM bitstuffing from 8bit to 7bit}
  function EightToSevenBit(const Input: string): string;

      {creates ASCII text from GSM coded text}
  function GsmToAscii(const Input: string; WantFormats: boolean = false): string;

      {converts ASCII chars to a string acc. to GSM 03.38}
  function AsciiToGsm(const Input: string): string;

      {calculates the length of the resulting string if it is coded}
  function AsciiToGsmLen(const Input: string; is7bit: boolean): integer;

      {calculates the length of the resulting string if it is coded}
  function AsciiLenForGsmTargetLen(const Input: string; const is7bit: boolean; const TargetLen: integer): integer;

      {creates 8bit data from a string with UCS characters}
  function SixteenToEightBit(const Input: widestring): string;

      {creates UCS characters from a 8bit string}
  function EightToSixteenBit(const Input: string): widestring;

      {creates a UCS2 string from a textfield in a SIM (11.11 Annex B)}
  function TextfieldToAscii(const Input: string): widestring;

      {creates a destination address IE from a string}
  function StringToBcdAddress0408(const Input: string): string;

{ conversions where the result is BCD }

      {creates a GSM BCD byte from a integer: 79 -> $97}
  function IntegerToBcd(const Input: integer): integer;

      {creates a GSM BCD byte from a string: '42' -> $24, '5' -> $F5}
  function StrToBcdByte(const Input: string): integer;

      {same as before, but with a complete string}
  function StrToBcdString(const Input: string): string;

{ conversions where the result is a single or multiple bytes }

      {creates a binary string from an integer, removes leading $00 bytes}
  function IntToBytes(const Input: cardinal): string;

      {creates a double byte string from a word}
  function WordToBytes(const Input: word): string;

      {creates a binary string from a hex coded string}
  function HexStringToString(const Input: string): string;
  function HexStringToWideString(const Input: string): widestring;

{ conversions where the result is some sort of text}

      {creates a string from a GSM BCD coded number, values > 9 will be omitted}
  function BcdNumberToString(const Input: string): string; // former BcdStringToHexStr

      {creates a string from a GSM BCD coded number, values > 14 will be omitted}
  function BcdStrToString(const Input: string): string;

      {creates a string from a GSM BCD coded number, values > 15 will be omitted}
  function BcdToString(const Input: string): string;

      {creates a string from a GSM BCD coded address}
  function BcdAddressToString(const Input: string): string;

      {creates a string of hexchars from a binary string}
  function BinToHexExt(const Input: string; const Spaces: boolean = true; const UpCase: boolean = true): string;
  function BinToHex(const Input: string): string;

{ conversions where the result is integer or cardinal}

      {reproducing the mathematical correct functions MOD and SMOD}
  function EMod(const x,y: integer): integer;
  function SMod(const x,y: integer): integer;

      {does a standard ORD, but returns 0 if Index > Length(From)}
  function OrdD(const From: string; const Index: integer): integer;     overload;
  function OrdD(const From: widestring; const Index: integer): integer; overload;

      {up to 4 bytes are converted to a cardinal, first byte is the MSB}
  function BytesToCardinal(const Input: string): cardinal;

      {gives you the bits from First to Last out of binary string Bits as an integer}
  function BitsToCardinal(const First,Last: integer; const Bits: string): cardinal;

      {gives the integer value of a GSM BCD number}
  function BcdNumberToInteger(const Input: char): integer;

{ conversions which does not fit in any other category}

      {expands a result like (0,1,3-5) in a complete set}
  function ExpandResultList(const List: string): string;

      {calculates the frequency from the GSM channel number}
  function ChannelToFreq(const Chan: integer; const IsDown: boolean): extended;

      {calculates the band type from a bcch number}
  function BcchToBandType(const Bcch: integer): TBandType;

      {calculates transmitted power in mW from TX_PWR}
  function PowerlevelToMilliwatt(const Pl,Bcch: integer): integer;

      {calculates RX_LEV from RSSI}
  function RssiToRxLevel(const Rssi: integer): integer;

      {extracts the BER-TLV-objects from a string}
  procedure SeparateTlvObjects(const Input: string; var TlvObjects: TTlvContainer);

      {checks if a string contains only #$FF}
  function IsEmptyEntry(const Entry: string): boolean;

      {makes a size string like '264 Byte' or '17,5 KByte' from a integer value}
  function SizeToText(const Size: integer): string;

const
  Numbers       = '-0123456789';
  NumberSetChar = ['0','1','2','3','4','5','6','7','8','9'];
  NumberSetNum  = [0,1,2,3,4,5,6,7,8,9];
  HexChars      = '0123456789ABCDEFabcdef';
  AddrChars     = '0123456789*#abc';
  CtrlChars     = #03#08#22#24;
  Gsm0705From   = '';
  Gsm0705To     = 'AAAAEEEIIIIOOOOUUUYaaaeeiiiooouuyy';
  GsmEscChars   = '^{}\[~]|'#27;
  GsmChars      = '@$'#10''#13'_ !"#%&'#39'()*+,-./0123456789:;<=>?ABCDEFGHIJKLMNOPQRSTUVWXYZܧabcdefghijklmnopqrstuvwxyz' + GsmEscChars;
  Band2G        : set of TBandType = [btGsm900,btEGsm,btGsm1800,btGsmR];
  Band3G        : set of TBandType = [btUmts];

implementation

constructor TTlvContainer.Create;
begin
  inherited;
  SetLength(FTlvArray,InitialTlvArraySize);
  Clear;
end;

destructor TTlvContainer.Destroy;
begin
  inherited;
end;

function TTlvContainer.GetTlvObject(i: integer): TTlvObject;
begin
  if i < FNextFree then Result := FTlvArray[i]
                   else Result := FTlvArray[FNextFree];
end;

function TTlvContainer.Count: integer;
begin
  Result := FNextFree;
end;

procedure TTlvContainer.Clear;
var
  i : integer;
begin
  FNextFree := 0;
  for i := 0 to High(FTlvArray) do
    begin
    FTlvArray[i].Tag   := 0;
    FTlvArray[i].Value := '';
    FTlvArray[i].Tlv   := '';
    end;
end;

procedure TTlvContainer.Add(const New: TTlvObject);
begin
  if FNextFree >= High(FTlvArray) then SetLength(FTlvArray,FNextFree + 10);
  FTlvArray[FNextFree].Tag   := New.Tag;
  FTlvArray[FNextFree].Value := New.Value;
  FTlvArray[FNextFree].Tlv   := New.Tlv;
  Inc(FNextFree);
end;

procedure TTlvContainer.Add(const New: string);
begin
  if FNextFree >= High(FTlvArray) then SetLength(FTlvArray,FNextFree + 10);
  FTlvArray[FNextFree].Tag   := OrdD(New,1);
  FTlvArray[FNextFree].Value := Copy(New,3,OrdD(New,2));
  FTlvArray[FNextFree].Tlv   := New;
  Inc(FNextFree);
end;

procedure TTlvContainer.Insert(const InsertAt: integer; const From: TTlvContainer);
var
  i : integer;
begin
  if FNextFree + From.Count >= High(FTlvArray) then SetLength(FTlvArray,FNextFree + From.Count + 10);
  for i := FNextFree - 1 downto InsertAt do
    FTlvArray[i + From.Count] := FTlvArray[i];
  for i := 0 to From.Count - 1 do
    FTlvArray[InsertAt + i] := From[i];
end;

function TTlvContainer.AllObjects: string;
var
  i : integer;
  s : string;
begin
  s := '';
  for i := 0 to FNextFree - 1 do s := s + FTlvArray[i].Tlv;
  Result := s;
end;

function FilterNumbers(const From: string): string;
var
  i : integer;
  s : string;
begin
  s := '';
  for i := 1 to Length(From) do if Pos(From[i],Numbers) > 0 then s := s + From[i];

  Result := s;
end;

function StrToBcdByte(const Input: string): integer;
var
  Lo,Hi,Value : integer;
begin
  if Length(Input) = 1 then Value := StrToIntDef('$' + Input + 'f',$FF)
                       else Value := StrToIntDef('$' + Input,$FF);
  Lo := Value and $0F;
  Hi := Value and $F0;
  Result := (Lo shl 4) or (Hi shr 4);
end;

function StrToBcdString(const Input: string): string;
var
  i,j : integer;
  s   : string;
begin
  i := 1;
  while i <= Length(Input) do
    begin
    j := StrToBcdByte(Copy(Input,i,2));
    if j <> $FF then s := s + Chr(j);
    i := i + 2;
    end;
  Result := s;
end;

function BcdNumberToInteger(const Input: char): integer;
begin
  Result := (Ord(Input) and $0F) * 10 + (Ord(Input) and $F0) shr 4;
end;

function ConvertBcdToString(const Input: string; const ValidValues: TValueRange): string;
var
  i   : integer;
  s   : string;
  Bcd1,Bcd2 : integer;
begin
  for i := 1 to Length(Input) do
    begin
    Bcd1 := Ord(Input[i]) and $0F;
    Bcd2 := (Ord(Input[i]) and $F0) shr 4;
    if Bcd1 in ValidValues then s := s + HexChars[Bcd1 + 1];
    if Bcd2 in ValidValues then s := s + HexChars[Bcd2 + 1];
    end;
  Result := s;
end;

function BcdNumberToString(const Input: string): string;
begin
  Result := ConvertBcdToString(Input,[0..9]);
end;

function BcdStrToString(const Input: string): string;
begin
  Result := ConvertBcdToString(Input,[0..14]);
end;

function BcdToString(const Input: string): string;
begin
  Result := ConvertBcdToString(Input,[0..15]);
end;

function BcdAddressToString(const Input: string): string;
var
  i   : integer;
  s   : string;
  Bcd1,Bcd2 : integer;
begin
  for i := 1 to Length(Input) do
    begin
    Bcd1 := Ord(Input[i]) and $0F;
    Bcd2 := (Ord(Input[i]) and $F0) shr 4;
    if Bcd1 in [0..14] then s := s + AddrChars[Bcd1 + 1];
    if Bcd2 in [0..14] then s := s + AddrChars[Bcd2 + 1];
    end;
  Result := s;
end;

function StringToBcdAddress0408(const Input: string): string;
var
  s : string;
begin
  if Copy(Input,1,1) = '+' then
    begin
    s := #$91 + StrToBcdString(Copy(Input,2,20));
    end
  else
    begin
    s := #$81 + StrToBcdString(Copy(Input,1,20));
    end;
  s := Chr(Length(s)) + s;
  s := s + StringOfChar(#$FF,12 - Length(s));
  Result := s;
end;

function IntegerToBcd(const Input: integer): integer;
begin
  Result := (Input mod 10) * 16 + Input div 10;
end;

function RssiToRxLevel(const Rssi: integer): integer;
begin
  Result := -110 + Rssi;
end;

function ChannelToFreq(const Chan: integer; const IsDown: boolean): extended;
begin
case Chan of
       0.. 124 : begin
                 Result := 890 + 0.2 * Chan;
                 if IsDown then Result := Result + 45;
                 end;
     512.. 885 : begin
                 Result := 1710.2 + 0.2 * (Chan - 512);
                 if IsDown then Result := Result + 95;
                 end;
     955..1023 : begin
                 Result := 890 + 0.2 * (Chan - 1024);
                 if IsDown then Result := Result + 45;
                 end;
     else        Result := 0;
     end;
end;

function BcchToBandType(const Bcch: integer): TBandType;
begin
  case Bcch of
    512.. 885 : Result := btGsm1800;
    0  .. 124 : Result := btGsm900;
    975..1023 : Result := btEGsm;
    955.. 974 : Result := btGsmR;
    else        Result := btUnknown;
    end;
end;

function PowerlevelToMilliwatt(const Pl,Bcch: integer): integer;
begin
  try
    if Pl < 45 then
      case BcchToBandType(Bcch) of
        btGsm1800 : Result := Trunc(Power(10,(30 - Pl * 2) * 0.1));
        btGsm900,
        btGsmR    : Result := Trunc(Power(10,(43 - Pl * 2) * 0.1));
        else        Result := 0;
      end
    else Result := MAXINT;
  except
    Result := 0;
  end;
end;

function SevenToEightBit(const Input: string): string;
var
  BitQueue : integer;
  NumBits  : integer;
  NextBits : integer;
  i        : integer;
begin
  BitQueue := 0;
  NumBits   := 0;
  for i := 1 to Length(Input) do
    begin
    NextBits := Ord(Input[i]);
    BitQueue := BitQueue or (NextBits shl NumBits);
    NumBits  := NumBits + 8;
    while NumBits > 6 do
      begin
      Result   := Result + Chr(BitQueue and $7F);
      BitQueue := BitQueue shr 7;
      NumBits  := NumBits - 7;
      end;
    end;
end;

function HexStringToString(const Input: string): string;
var
  Hex, Output: string;
  i,j        : integer;
begin
  Hex := StringOfChar(#0,Length(Input));
  i := 1;
  j := 1;
  while i <= Length(Input) do
    begin
    if Pos(Input[i], HexChars) > 0 then
      begin
      Hex[j] := Input[i];
      Inc(j);
      end;
    Inc(i);
    end;
  Hex := Trim(Hex);
  i := 1;
  if Length(Hex) > 1 then
    repeat
    Output := Output + Chr(StrToInt('$' + Copy(Hex,i,2)));
    Inc(i,2);
    until i > Length(Hex);
  Result := Output;
end;

function HexStringToWideString(const Input: string): widestring;
var
  Hex    : string;
  Output : widestring;
  i,j    : integer;
begin
  Hex := StringOfChar(#0,Length(Input));
  i := 1;
  j := 1;
  while i <= Length(Input) do
    begin
    if Pos(Input[i], HexChars) > 0 then
      begin
      Hex[j] := Input[i];
      Inc(j);
      end;
    Inc(i);
    end;
  Hex := Trim(Hex);
  i := 1;
  if Length(Hex) > 1 then
    repeat
    Output := Output + WideChar(StrToInt('$' + Copy(Hex,i,4)));
    Inc(i,4);
    until i > Length(Hex);
  Result := Output;
end;

function IntToBytes(const Input: cardinal): string;
var
  Temp : string;
begin
  Temp := IntToHex(Input,8);
  while Copy(Temp,1,2) = '00' do Delete(Temp,1,2);
  Result := HexStringToString(Temp);
end;

function WordToBytes(const Input: word): string;
begin
  Result := HexStringToString(IntToHex(Input,4));
end;

function EightToSevenBit(const Input: string): string;
var
  BitQueue : integer;
  NumBits  : integer;
  NextBits : integer;
  i        : integer;
begin
  BitQueue := 0;
  NumBits  := 0;
  for i := 1 to Length(Input) do
    begin
    NextBits := Ord(Input[i]) and $7F;
    BitQueue := BitQueue or (NextBits shl NumBits);
    NumBits  := NumBits + 7;
    if NumBits > 7 then
       begin
       Result   := Result + Chr(BitQueue and $FF);
       BitQueue := BitQueue shr 8;
       NumBits  := NumBits - 8;
       end;
    end;
  if NumBits > 0 then Result := Result + Chr(BitQueue and $FF);
end;

function GsmToAscii(const Input: string; WantFormats: boolean = false): string;
const     //0         1         2         3            4         5         6         7         8         9         0
Ascii    = '@$Ƿ _           !"#%&''()*+,-./0123456789:;<=>?ABCDEFGHIJKLMNOPQRSTUVWXYZܧabcdefghijklmnopqrstuvwxyz';
Extended = '@$Ƿ _  ^        !"#%&''{}*+,-.\0123456789:;[~]?|ABCDEFGHIJKLMNOPQRSTUVWXYZܧabcdfghijklmnopqrstuvwxyz';

var
  r : string;
  a : integer;
begin
  a := 1;
  while a <= Length(Input) do
    begin
    case Input[a] of
      #010       : if WantFormats then r := r + #10 else r := r + ' ';
      #013       : if WantFormats then r := r + #13 else r := r + ' ';
      #027       : if a < Length(Input) then
                     begin
                     Inc(a);
                     r := r + Extended[Ord(Input[a])+1];
                     end else r := r + ' ';
      #128..#255 : ;
      else r := r + Ascii[Ord(Input[a])+1];
      end;
    Inc(a);
    end;
Result := r;
end;

function TextfieldToAscii(const Input: string): widestring;
var
  CodingType : integer;
  s          : widestring;
  i,j,Offset : integer;
begin
  CodingType := OrdD(Input,1);
  case CodingType of
    $00..$7F : begin
               s := GsmToAscii(Input,false);
               end;
    $80      : begin
               for i := 1 to Length(Input) div 2 do
                 begin
                 j := BytesToCardinal(Copy(Input,2*i,2));
                 if j <> $FFFF then s := s + WideChar(j);
                 end;
               end;
    $81      : begin
               Offset := OrdD(Input,3) shl 7;
               for i := 4 to OrdD(Input,2) + 3 do
                 if Length(Input) <= i then
                   begin
                   j := Ord(Input[i]);
                   if j > $7F then s := s + WideChar(Offset + (j and $7F))
                              else s := s + WideString(GsmToAscii(Input[i],false));
                   end;
               end;
    $82      : begin
               Offset := BytesToCardinal(Copy(Input,3,2));
               for i := 4 to OrdD(Input,2) + 3 do
                 if Length(Input) <= i then
                   begin
                   j := Ord(Input[i]);
                   if j > $7F then s := s + WideChar(Offset + (j and $7F))
                              else s := s + WideString(GsmToAscii(Input[i],false));
                   end;
               end;
    else       s := '';
    end;
  Result := s;
end;

function BinToHexExt(const Input: string; const Spaces: boolean = true; const UpCase: boolean = true): string;
var
  Loop      : integer;
  HexResult : string;
begin
  if Spaces then SetLength(HexResult,Length(Input) * 3)
            else SetLength(HexResult,Length(Input) * 2);
  if Spaces then
    for Loop := 1 to Length(Input) do
      begin
      HexResult[Loop*3-2] := HexChars[((Ord(Input[Loop]) and $F0) shr 4) + 1];
      HexResult[Loop*3-1] := HexChars[ (Ord(Input[Loop]) and $0F)        + 1];
      HexResult[Loop*3]   := ' ';
      end
  else HexResult := BinToHex(Input);
  if UpCase then Result := AnsiUpperCase(HexResult)
            else Result := AnsiLowerCase(HexResult);
end;

function BinToHex(const Input: string): string;
var
  Loop : integer;
  s    : string;
begin
  SetLength(s,Length(Input) * 2);
  for Loop := 1 to Length(Input) do
    begin
    s[Loop*2-1] := HexChars[((Ord(Input[Loop]) and $F0) shr 4) + 1];
    s[Loop*2]   := HexChars[ (Ord(Input[Loop]) and $0F)        + 1];
    end;
  Result := s;
end;

function OrdD(const From: string; const Index: integer): integer;
begin
  if (Index < 1) or
     (Index > Length(From)) then Result := 0
                            else Result := Ord(From[Index]);
end;

function OrdD(const From: widestring; const Index: integer): integer;
begin
  if (Index < 1) or
     (Index > Length(From)) then Result := 0
                            else Result := Ord(From[Index]);
end;

function BytesToCardinal(const Input: string): cardinal;
begin
  case Length(Input) of
    0  : Result := 0;
    1  : Result := Ord(Input[1]);
    2  : Result := (Ord(Input[1]) shl 8) + Ord(Input[2]);
    3  : Result := (Ord(Input[1]) shl 16) + (Ord(Input[2]) shl 8) + Ord(Input[3]);
    else Result := (Ord(Input[1]) shl 24) + (Ord(Input[2]) shl 16) + (Ord(Input[3]) shl 8) + Ord(Input[4]);
    end;
end;

function BitsToCardinal(const First,Last: integer; const Bits: string): cardinal;
var
  NumBytes,RawBits,BitMask,StartByte,EndByte : integer;
begin
  StartByte := Length(Bits) - ((First - 1) div 8);
  EndByte   := Length(Bits) - ((Last - 1) div 8);
  NumBytes  := EndByte - StartByte + 1;
  if NumBytes < 5 then
    begin
    BitMask := (1 shl (First - Last + 1)) - 1;
    RawBits := BytesToCardinal(Copy(Bits,StartByte,NumBytes));
    RawBits := RawBits shr ((Last - 1) mod 8);
    Result := RawBits and BitMask;
    end
  else Result := 0;
end;

function EMod(const x,y: integer): integer;
begin
  Result := x mod y;
  if Result < 0 then Result := y + Result;
end;

function SMod(const x,y: integer): integer;
begin
  Result := EMod(x,y);
  if Result = 0 then Result := y;
end;

function IsEmptyEntry(const Entry: string): boolean;
begin
  if (Entry = StringOfChar(#$FF,Length(Entry))) or
     (Entry = ''                              ) or
     (Entry = StringOfChar(#$00,Length(Entry)))    then Result := true
                                                   else Result := false;
end;

procedure SeparateTlvObjects(const Input: string; var TlvObjects: TTlvContainer);
var
  i   : integer;
  Tlv : TTlvObject;
begin
  if Length(Input) > 1 then
    begin
    i := 1;
    while i < Length(Input) do
//      if Pos(Input[i],#$00#$FF) = 0 then
      if Pos(Input[i],#$FF) = 0 then
        begin
        Tlv.Tag   := Ord(Input[i]);
        Tlv.Value := Copy(Input,i + 2,Ord(Input[i+1]));
        Tlv.Tlv   := Copy(Input,i,Ord(Input[i+1]) + 2);
        TlvObjects.Add(Tlv);
        Inc(i,Ord(Input[i+1])+2);
        end else Inc(i);
    end;
end;

function SixteenToEightBit(const Input: widestring): string;
var
  s : string;
  a : integer;
begin
  SetLength(s,Length(Input) * 2);
  for a := 1 to Length(Input) do
    begin
    s[a*2-1] := Chr(Ord(Input[a]) div 256);
    s[a*2]   := Chr(Ord(Input[a]) mod 256);
    end;
  Result := s;
end;

function EightToSixteenBit(const Input: string): widestring;
var
  a,b : integer;
  w   : widestring;
  c   : widechar;
begin
  for a := 1 to Length(Input) div 2 do
    begin
    b := a * 2 - 1;
    c := WideChar((OrdD(Input,b) shl 8) + OrdD(Input,b + 1));
    w := w + c;
    end;
  Result := w;
end;

function AsciiToGsm(const Input: string): string;
const //                                    10                                      20                                      30
a2g = #032#032#032#032#032#032#032#032#032#010#032#032#013#032#032#032#032#032#032#032#032#032#032#032#032#032#032#032#032#032#032#032 + //32
      #033#034#035#002#037#038#039#040#041#042#043#044#045#046#047#048#049#050#051#052#053#054#055#056#057#058#059#060#061#062#063#000 + //64
      #065#066#067#068#069#070#071#072#073#074#075#076#077#078#079#080#081#082#083#084#085#086#087#088#089#090#188#175#190#148#017#032 + //96
      #097#098#099#100#101#102#103#104#105#106#107#108#109#110#111#112#113#114#115#116#117#118#119#120#121#122#168#192#169#189#032#229 + //128
      #032#032#032#032#032#032#032#032#032#032#032#032#032#032#032#032#032#032#032#032#032#032#032#032#032#032#032#032#032#032#032#032 + //160
      #064#032#001#036#003#032#095#032#032#032#032#032#032#032#032#032#032#032#032#032#032#032#013#032#032#032#032#032#032#032#096#032 + //192
      #032#032#032#091#014#028#009#032#031#032#032#032#032#032#032#032#093#032#032#032#032#092#032#011#032#032#032#094#032#032#030#127 + //224
      #032#032#032#123#015#029#032#004#005#032#032#007#032#032#032#032#125#008#032#032#032#124#032#012#006#032#032#126#032#032#032#032;  //256
var
  r : string;
  c : char;
  a : integer;
begin
  a := 1;
  while a <= Length(Input) do
    begin
    c := a2g[Ord(Input[a])];
    case c of
      #000..#127 : r := r + c;
      #128..#255 : r := r + #27 + Chr(Ord(c)-128);
      end;
    Inc(a);
    end;
  Result := r;
end;

function AsciiToGsmLen(const Input: string; is7bit: boolean): integer;
const
  EscChar = ['C','E','L','0','a'..'z'];
var
  c   : char;
  r,a : integer;
begin
  r := 0;
  a := 1;
  while a <= Length(Input) do
    begin
    c := Input[a];
    if c = '\' then
      begin
      Inc(a);
      if a <= Length(Input) then
        begin
        c := Input[a];
        if c in EscChar then Inc(r)
          else if c = 'P' then Inc(r,2)
            else if Pos(c,GsmEscChars) > 0 then Inc(r,2)
              else if Pos(c,GsmChars) > 0 then Inc(r);
        end;
      end
    else
      begin
      if is7bit then
        begin
        if (Pos(c,GsmChars) > 0) or
           (Pos(c,Gsm0705From) > 0) then Inc(r);
        if Pos(c,GsmEscChars) > 0 then Inc(r);
        end
      else
        begin
        Inc(r);
        end;
      end;
    Inc(a);
    end;
  Result := r;
end;

function AsciiLenForGsmTargetLen(const Input: string; const is7bit: boolean; const TargetLen: integer): integer;
const
  EscChar = ['C','E','L','0','\','a'..'z'];
var
  c   : char;
  i,o : integer;
begin
  o := 0;
  i := 0;
  while (i < Length(Input)) and (o < TargetLen) do
    begin
    Inc(i);
    c := Input[i];
    if c = '\' then
      begin
      Inc(i);
      if i <= Length(Input) then
        begin
        c := Input[i];
        if c in EscChar then Inc(o)
          else if c = 'P' then Inc(o,2)
            else if Pos(c,GsmEscChars) > 0 then Inc(o,2)
              else if Pos(c,GsmChars) > 0 then Inc(o);
        end;
      end
    else
      begin
      if is7bit then
        begin
        if (Pos(c,GsmChars) > 0) or
           (Pos(c,Gsm0705From) > 0) then Inc(o);
        if Pos(c,GsmEscChars) > 0 then Inc(o);
        end
      else
        begin
        Inc(o);
        end;
      end;
    end;
  if o > TargetLen then Dec(i);
  Result := i;
end;

function ExpandResultList(const List: string): string;
var
  v           : array[0..99] of string;
  c,i,j,p     : integer;
  First, Last : integer;
begin
  c := SortOutSubstrings(List,v,[','],false);
  for i := 0 to c - 1 do
    begin
    p := Pos('-',v[i]);
    if p > 0 then
      begin
      First := StrToIntDef(Copy(v[i],1,p - 1),MAXINT);
      Last  := StrToIntDef(Copy(v[i],p + 1,Length(v[i]) - p),MAXINT);
      if First > Last then
        begin
        p := First;
        First := Last;
        Last := p;
        end;
      v[i] := '';
      for j := First to Last do
        begin
        if v[i] <> '' then v[i] := v[i] + ',';
        v[i] := v[i] + IntToStr(j);
        end;
      end;
    end;
  Result := '';
  for i := 0 to c - 1 do
    begin
    if Result <> '' then Result := Result + ',';
    Result := Result + v[i];
    end;
end;

function SizeToText(const Size: integer): string;
begin
  case Size of
  0..1023           : Result := Format('%d Byte',[Size]);
  1024..1024*1024-1 : Result := Format('%.1f KByte',[Size/1024]);
  else                Result := Format('%.1f MByte',[Size/(1024*1024)]);
  end;
end;

end.
