{******************************************************************}
{                                                                  }
{ PC/SC Interface component                                        }
{ Helps you access processor cards through the MS SmartCard API    }
{                                                                  }
{ The main focus for this component is reading GSM/3GPP SIMs, but  }
{ other cards should, in general, work too                         }
{                                                                  }
{ The Original Code is PCSCConnector.pas                           }
{                                                                  }
{ The Initial Developer of the Original Code is                    }
{ Norbert Huettisch (nobbi(at)nobbi.com)                           }
{                                                                  }
{ Any suggestions and improvements to the code are appreciated     }
{                                                                  }
{ This Code uses a modified   SCardErr.pas (included)              }
{ This Code uses a modified   WinSCard.pas (included)              }
{ This code uses the original WinSmCrd.pas (included)              }
{                                                                  }
{ All originally made by Chris Dickerson (chrisd(at)tsc.com),      }
{ available as 'Interface units for the Microsoft Smart Card API'  }
{ at the Project JEDI Homepage http://www.delphi-jedi.org          }
{                                                                  }
{                                                                  }
{******************************************************************}
{                                                                  }
{ The contents of this file are subject to the                     }
{                                                                  }
{       Mozilla Public License Version 1.1 (the "License")         }
{                                                                  }
{ You may not use this file except in compliance with the License. }
{ You may obtain a copy of the License at
{                                                                  }
{                   http://www.mozilla.org/MPL/                    }
{                                                                  }
{ Software distributed under the License is distributed on an      }
{ "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either expressed or }
{ implied. See the License for the specific language governing     }
{ rights and limitations under the License.                        }
{                                                                  }
{******************************************************************}

{$RANGECHECKS OFF}

unit PCSCConnector;

interface

uses
  Windows, Messages, Forms, Classes, SysUtils, StrUtils,
  SCardErr, WinSCard, WinSmCrd, GsmHelper, Utilities;

type
  TInitType          = (itGSimFirst,itUSimFirst,itGSimOnly,itUSimOnly);
  TErrSource         = (esInit, esConnect, esGetStatus, esTransmit);
  TCardClass         = (ccUnknown,   // the type of card is unknown
                        ccUicc,      // a UICC conforming to TS 102221
                        ccGsmSim,    // a GSM SIM
                        ccJanus,     // a GSM SIM with applications from EFdir
                        ccUSim,      // a UICC with a USIM application
                        ccTSim,      // a UICC with a TETRA application
                        ccOtherApp); // a UICC with a unknown application
  TCardVoltage       = (vcUnknown,vc50Volt,vc30Volt,vc18Volt,vcOtherClass);
  TNeededPIN         = (npPIN1, npPIN2, npPUK1, npPUK2, npUPIN);
  TChvState          = (csNotAvailable,csNotPresented,csChvOk,csChvNotOk,csBlocked,csChvRefused,csChvDisabled);
  TFileStructure     = (fsNone,fsUnknown,fsTransparent,fsLinearFixed,fsCyclic);
  TFileType          = (ftUnknown,ftReserved,ftMaster,ftDedicated,ftElementary);
  TAccessCondition   = (acUnknown,acAlways,acChv1,acChv2,acChvU,acRfu,acAdm,acNever,acProt,acExtAuth);
  TClockStopCond     = (csNotAllowed,csNoPref,csHighPref,csLowPref,csOnlyHigh,csOnlyLow);
  TDelimiters        = set of Char;

  TOnErrorEvent      = procedure(Sender: TObject; ErrSource: TErrSource; ErrCode: cardinal) of object;
  TOnNeedPinEvent    = procedure(Sender: TObject; NeedPIN: TNeededPIN) of object;
  TOnLogDataEvent    = procedure(Sender: TObject; Location,Status,Data: string) of object;
  TOnProgressEvent   = procedure(Sender: TObject; Accessed: string; RecNo: integer) of object;

const
  MAXAPDULENGTH      = 260; // CLA + INS + P1..3 + 255Bytes
  NOREADERSELECTED   = -1;
  SCARD_PCI_T0       : SCARD_IO_REQUEST = (dwProtocol:1; dbPciLength:8);
  SCARD_PCI_T1       : SCARD_IO_REQUEST = (dwProtocol:2; dbPciLength:8);
  SCARD_PROTOCOL_T0  = $00000001;
  SCARD_PROTOCOL_T1  = $00000002;
  SCARD_PROTOCOL_RAW = $00010000;
  SCARD_PROTOCOL_UNK = $00000000;

  WM_CARDSTATE     = WM_USER + 42;

  SCARD_E_PCSC_NOTLOADED  = cardinal($E0000001);   //error severity, user flag, system facility, error #1
  SCARD_E_INVALID_CONTEXT = cardinal($E0100002);   //error severity, user flag, card facility, error #2

  SIM_STATE_OK             = $9000;
  SIM_STATE_HAVEDATA       = $9F;
  SIM_STATE_DOGETRESPONSE  = $61;
  SIM_STATE_WRONGSTRUCTURE = $6981;
  SIM_STATE_P1P2INCORRECT  = $6B00;
  SIM_STATE_WRONGCLASS     = $6E00;
  SIM_STATE_TECHPROBLEMS   = $6F00;
  SIM_STATE_FILENOTFOUND   = $9404;
  SIM_STATE_ACCESSDENIED   = $9804;
  SIM_STATE_CHVBLOCKED     = $9840;

  TAGEtsiFcp               = $62;
  TAGFileDesc              = $82;
  TAGFileId                = $83;
  TAGAid                   = $84;
  TAGPropInfo              = $A5;
  TAGLifeCycle             = $8A;
  TAGSecAttribA            = $86;
  TAGSecAttribB            = $8B;
  TAGSecAttribCompact      = $8C;
  TAGSecAttribExpanded     = $AB;
  TAGPinStatus             = $C6;
  TAGFileSize              = $80;
  TAGTotFileSize           = $81;

  TAGKeyReference          = $83;
  TAGUsageQualifier        = $95;

  GsmPin1                  = #$01;
  GsmPin2                  = #$02;
  IccPin1                  = #$01;
  IccPin2                  = #$81;
  IccUPin                  = #$11;

  AccRuleRead              = $01;
  AccRuleUpdate            = $02;
  AccRuleWriteB            = $04;
  AccRuleDeactivate        = $08;
  AccRuleActivate          = $10;
  AccRuleXXX               = $20;
  AccRuleYYY               = $40;
  AccRuleIncrease          = $32;

  AccAlways                = cardinal($0000);
  NoAccess                 = cardinal($0001);
  NeedPin1                 = cardinal($0002);
  NeedPin2                 = cardinal($0004);
  NeedPinU                 = cardinal($0008);
  NeedAdm5                 = cardinal($0010);
  NeedAdm6                 = cardinal($0020);
  OrRelation               = cardinal($8000);

type
  TFileInfo = record
              FileType    : TFileType;
              FileId      : string;
              AccessRules : string;
              // for a EFile
                FileStruct  : TFileStructure;
                FileSize    : integer;
                RecordLen   : integer;
                NumRecords  : integer;
                ReadCond    : TAccessCondition;
                WriteCond   : TAccessCondition;
              // for a DFile
                AppId       : string;
                BytesFree   : integer;
                NumDFiles   : integer;
                NumEFiles   : integer;
                NumCodes    : integer;
                Chv1Enable  : boolean;
                Chv1Tries   : integer;
                Chv2Tries   : integer;
                Chv1UTries  : integer;
                Chv2UTries  : integer;
                ClockStop   : TClockStopCond;
              end;

  TAccessBitmap  = cardinal;

  TAccessRule    = record
                   ReadAccess  : TAccessBitmap;
                   WriteAccess : TAccessBitmap;
                   end;

  TAccessRules   = array of TAccessRule;

  TPCSCConnector = class(TObject)
  protected
    FDebugging          : boolean;
    FContext            : integer;
    FCardHandle         : integer;
    FConnected          : boolean;
    FNumReaders         : integer;
    FUseReaderNum       : integer;
    FReaderList         : TStringlist;
    FAttrProtocol       : integer;
    FAttrICCType        : string;
    FAttrCardATR        : string;
    FAttrVendorName     : string;
    FAttrVendorSerial   : string;
    FCurrentFile        : string;
    FLastStatus         : string;
    FLastFileInfo       : TFileInfo;
    FAccessRuleFile     : string;
    FAccessRules        : TAccessRules;
    FChv1State          : TChvState;
    FChv2State          : TChvState;
    FChvUState          : TChvState;
    FCardVoltage        : TCardVoltage;
    FCardClass          : TCardClass;
    FExpectCardClass    : TCardClass;

    FOnReaderWaiting    : TNotifyEvent;
    FOnReaderListChange : TNotifyEvent;
    FOnCardInserted     : TNotifyEvent;
    FOnCardActive       : TNotifyEvent;
    FOnCardRemoved      : TNotifyEvent;
    FOnCardInvalid      : TNotifyEvent;
    FOnError            : TOnErrorEvent;
    FOnCHVNeeded        : TOnNeedPinEvent;
    FOnLogMessage       : TOnLogDataEvent;
    FOnProgress         : TOnProgressEvent;

    procedure SetReaderNum(Value: integer);
    procedure MessageWndProc(var Msg: TMessage);
    function  ConnectSelectedReader: boolean;
    procedure ProcessReaderState(const OldState,NewState: cardinal);
    procedure GetReaderAttributes;
    procedure GetCardAttributes;
    procedure ClearReaderAttributes;
    procedure ClearCardAttributes;
    procedure CheckIfPcscIsLoaded;
    function  IsReaderOpen: boolean;
    function  GetReaderState: cardinal;
    procedure CloseAndDisconnect;
    procedure CardInsertedAction;
    procedure CardActiveAction;
    procedure CardRemovedAction;
    procedure ChangeCardClass(const NewClass: TCardClass);

  public
    constructor Create;
    destructor  Destroy; override;
    function    Init: boolean;
    function    Open: boolean;
    procedure   Close;
    function    Connect: boolean;
    procedure   Disconnect;
    function    CardIsValidContext(Context: cardinal): cardinal;
// Layer 1
    function    SendPduToCard(const Apdu: string): string;
    function    SendCommandToCard(const Command: string; var Data: string; var Status: word): boolean;
// Layer 2 should possibly be moved to the private section
    function    CardCommandStatus(var Status: string): word;
    function    CardCommandSelect(const FileID: string; var Status: string): word;
    function    CardCommandReadBinary(const Offset, Length: integer; var Data: string): word;
    function    CardCommandReadRecord(const RecordNum, Length: integer; var Data: string): word;
    function    CardCommandReadCurrentRecord(const Length: integer; var Data: string): word;
    function    CardCommandReadNextRecord(const Length: integer; var Data: string): word;
    function    CardCommandUpdateBinary(const Offset: integer; var Data: string): word;
    function    CardCommandUpdateRecord(const RecordNum: integer; var Data: string): word;
    function    CardCommandVerifyChv(const ChvType: TNeededPIN; const ChvValue: string): word;
    function    CardCommandChangeChv(const ChvType: TNeededPIN; const OldValue,NewValue: string): word;
    function    CardCommandUnblockChv(const ChvType: TNeededPIN; const UnblockChv,NewChv: string): word;
// Layer 3
    function    CardCheckAccess(const Needed: TAccessCondition): boolean;
    function    CardReadFile(const FileID: string; var Status,Data: string): word;
    function    CardReadTransparentFile(const FileID: string; var Status,Data: string): word;
    function    CardSelectMF(var Status: string): word;
    function    CardChangeDir(const FileID: string; var Status: string): word; overload;
    function    CardChangeDir(const FileID: word; var Status: string): word;   overload;
    function    CardReadRecordsTopDown(const FileID: string; var Status,Data: string): word;
    function    CardReadRecordsCyclic(const FileID: string; var Status,Data: string): word;
    function    CardUpdateRecord(const FileID: string; const RecordNum: integer; var Data: string): word;
    function    CardUpdateBinary(const FileID: string; const Offset: integer; var Data: string): word;
    function    CardVerifyChv(const ChvType: TNeededPIN; const ChvValue: string): word;
    function    CardChangeChv(const ChvType: TNeededPIN; const OldValue,NewValue: string): word;
    function    CardUnblockChv(const ChvType: TNeededPIN; const UnblockChv,NewValue: string): word;
// Higher Layers
    procedure   DecodeFileStatus(const Status: string; var FileInfo: TFileInfo);
    procedure   ClearFileInfo(var FileInfo: TFileInfo);
    function    DecodeResponseCode(const Code: word): string;
    function    DecodeSCardReturnCode(const Code: cardinal): string;

  published
    property UseReaderNum:       integer          read FUseReaderNum       write SetReaderNum  default -1;

    property OnCardInserted:     TNotifyEvent     read FOnCardInserted     write FOnCardInserted;
    property OnCardActive:       TNotifyEvent     read FOnCardActive       write FOnCardActive;
    property OnCardRemoved:      TNotifyEvent     read FOnCardRemoved      write FOnCardRemoved;
    property OnCardInvalid:      TNotifyEvent     read FOnCardInvalid      write FOnCardInvalid;
    property OnReaderWaiting:    TNotifyEvent     read FOnReaderWaiting    write FOnReaderWaiting;
    property OnReaderListChange: TNotifyEvent     read FOnReaderListChange write FOnReaderListChange;
    property OnError:            TOnErrorEvent    read FOnError            write FOnError;
    property OnChvNeeded:        TOnNeedPinEvent  read FOnChvNeeded        write FOnChvNeeded;
    property OnLogMessage:       TOnLogDataEvent  read FOnLogMessage       write FOnLogMessage;
    property OnProgress:         TOnProgressEvent read FOnProgress         write FOnProgress;

    property ReaderList:       TStringList  read FReaderList;
    property NumReaders:       integer      read FNumReaders;
    property Connected:        boolean      read FConnected;
    property Opened:           boolean      read IsReaderOpen;
    property ReaderState:      cardinal     read GetReaderState;
    property AttrProtocol:     integer      read FAttrProtocol;
    property AttrICCType:      string       read FAttrICCType;
    property AttrCardATR:      string       read FAttrCardATR;
    property AttrVendorName:   string       read FAttrVendorName;
    property AttrVendorSerial: string       read FAttrVendorSerial;
    property CurrentFile:      string       read FCurrentFile;
    property LastStatus:       string       read FLastStatus;
    property LastFileInfo:     TFileInfo    read FLastFileInfo;
    property AccessRules:      TAccessRules read FAccessRules;
    property Chv1State:        TChvState    read FChv1State               write FChv1State;
    property Chv2State:        TChvState    read FChv2State               write FChv2State;
    property ChvUState:        TChvState    read FChvUState               write FChvUState;
    property CardVoltage:      TCardVoltage read FCardVoltage;
    property CardClass:        TCardClass   read FCardClass               write ChangeCardClass;
    property ExpectCardClass:  TCardClass   read FExpectCardClass         write FExpectCardClass;

    property Debugging:        boolean      read FDebugging               write FDebugging;

  end;

implementation

var
  ActReaderState  : cardinal;
  LastReaderState : cardinal;
  SelectedReader  : PChar;
  ReaderOpen      : boolean;
  NotifyHandle    : HWND;
  SystemInfo      : tosVersionInfo;
  PCSCLoaded      : boolean;
  CmdGetStatus    : string;
  CmdGetResponse  : string;
  CmdSelectFile   : string;
  CmdReadBinary   : string;
  CmdReadRecord   : string;
  CmdReadCurrRec  : string;
  CmdReadNextRec  : string;
  CmdVerifyChv    : string;
  CmdChangeChv    : string;
  CmdUnblockChv   : string;
  CmdUpdateBinary : string;
  CmdUpdateRecord : string;

const
  // GSM Command strings
  GsmCmdGetStatus      = #$A0#$F2#$00#$00#$16;
  GsmCmdGetResponse    = #$A0#$C0#$00#$00;
  GsmCmdSelectFile     = #$A0#$A4#$00#$00#$02;
  GsmCmdReadBinary     = #$A0#$B0;
  GsmCmdReadRecord     = #$A0#$B2;
  GsmCmdReadCurrRecord = #$A0#$B2#$00#$04;
  GsmCmdReadNextRecord = #$A0#$B2#$00#$02;
  GsmCmdUpdateBinary   = #$A0#$D6;
  GsmCmdUpdateRecord   = #$A0#$DC;
  GsmCmdVerifyChv      = #$A0#$20#$00;
  GsmCmdChangeChv      = #$A0#$24#$00;
  GsmCmdUnblockChv     = #$A0#$2C#$00;

  // USIM Command strings
  UsimCmdGetStatus      = #$80#$F2#$00#$00;
  UsimCmdGetResponse    = #$00#$C0#$00#$00;
  UsimCmdSelectFile     = #$00#$A4#$00#$04#$02;
  UsimCmdSelectMF       = #$00#$A4#$00#$0C#$00;
  UsimCmdSelectAppl     = #$00#$A4#$04#$04;
  UsimCmdReadBinary     = #$00#$B0;
  UsimCmdReadRecord     = #$00#$B2;
  UsimCmdReadCurrRecord = #$00#$B2#$00#$04;
  UsimCmdReadNextRecord = #$00#$B2#$00#$02;
  UsimCmdUpdateBinary   = #$00#$D6;
  UsimCmdUpdateRecord   = #$00#$DC;
  UsimCmdVerifyChv      = #$00#$20#$00;
  UsimCmdChangeChv      = #$00#$24#$00;
  UsimCmdUnblockChv     = #$00#$2C#$00;

// STATUS "F2" "00" "00" lgth R

  DFmaster      = word($3F00);
  EFdir         = word($2F00);
  DFgsm         = word($7F20);
  DFtetra       = word($7f90);

////////////////////////////////////////////////////////////////////////////////
// local procedures come in here

{ SCardErr defines all error vaues as DWORD, but all functions from WinSCard
are returning a LONGINT. Because this may cause a 'range check' error,
the longint has to be converted to a dword. Done this here.  }

function LongintToDword(const i: longint): dword;
begin
  Move(i,Result,SizeOf(Result));
end;

function CardWatcherThread(PContext: pointer): integer;
var
  RetVar   : cardinal;
  RContext : cardinal;
  RState   : array[0..1] of SCARD_READERSTATEA;
begin
  try
    RContext := cardinal(PContext^);
    FillChar(RState,SizeOf(RState),#0);
    RState[0].szReader     := SelectedReader;
    RState[0].pvUserData   := nil;
    RState[0].dwEventState := ActReaderState;
    while ReaderOpen do
      begin
      RState[0].dwCurrentState := RState[0].dwEventState;
      RetVar := LongintToDword(SCardGetStatusChange(RContext, -1, RState, 1));
      ActReaderState := RState[0].dwEventState;
      PostMessage(NotifyHandle, WM_CARDSTATE, RetVar, 0);
      end;
  finally
    Result := 0;
  end;
end;

function DecodeCommandFromPdu(const Pdu: string): string;
begin
  try
    case Pdu[2] of
      #$20 : Result := 'VERIFY CHV     ';
      #$A4 : Result := 'SELECT FILE    ';
      #$B0 : Result := 'READ BINARY    ';
      #$B2 : Result := 'READ RECORD    ';
      #$C0 : Result := 'GET RESPONSE   ';
      #$D6 : Result := 'UPDATE BINARY  ';
      #$DC : Result := 'UPDATE RECORD  ';
      #$F2 : Result := 'GET STATUS     ';
      #$24 : Result := 'CHANGE CHV     ';
      else   Result := 'UNKNOWN COMMAND';
      end;
  finally end;
end;

function DecodeCardClass(const CardClass: TCardClass): string;
begin
  case CardClass of
    ccUnknown  : Result := 'Unknown';
    ccUicc     : Result := 'UICC';
    ccGsmSim   : Result := 'GSM SIM';
    ccJanus    : Result := 'JANUS SIM';
    ccUSim     : Result := 'USIM';
    ccTSim     : Result := 'TSIM';
    ccOtherApp : Result := 'UICC with App';
    else         Result := 'Undefined';
    end;
  Result := Result + ' (' + IntToStr(Ord(CardClass)) + ')';
end;

function KeyrefUsequalToAccessBitmap(const ScDo: string; const Preload: cardinal): TAccessBitmap;
var
  i,j : integer;
begin
  Result := Preload;
  i := 1;
  while (i + 5) <= Length(ScDo) do
    begin
    case OrdD(ScDo,i) of
      TAGKeyReference : begin
                        if (OrdD(ScDo,i + 3) = TAGUsageQualifier) and (OrdD(ScDo,i + 5) = $08) then
                          begin
                          j := OrdD(ScDo,i + 2);
                          case j of
                            $01      : Result := Result or NeedPin1;
                            $81      : Result := Result or NeedPin2;
                            $11      : Result := Result or NeedPinU;
                            $0A..$0E : Result := Result or NeedAdm5;
                            $8A..$8E : Result := Result or NeedAdm6;
                            end;
                          end;
                        Inc(i,6);
                        end;
      $A4             : begin
                        Inc(i,2);
                        end;
      else              Inc(i);
      end;
    end;
end;

{procedure DecodeExpandedAccessRule(const Rule: string; var R,W: TAccessBitmap);

The content of file 2F06 is as referenced in TS102221 - Annex E

  PIN1 is : Application PIN  - PIN        Appl.1 global (key reference '01');
  PIN2 is : Application PIN2 - Second PIN Appl.1 local  (key reference '81');
  UPIN is : Universal PIN    - UNIVERSAL PIN            (key reference '11');
  ADM  is :                    ADM5                     (key reference '0A'..'0E' and
                               ADM6                                    '8A'..'8E')
ARR format:

 ARR         : (AM_TLV) (SC_TLV)
 AM_TLV      : 80    READ > 01  UPDATE > 02
 SC_TLV      : 90    ALWAYS
               A4    AND Template (KEYREF_TLV) (USEQUAL_TLV)
               A0    OR Template
               97    NEVER
 KEYREF_TLV  : 83    key reference as stated above
 USEQUAL_TLV : 95    08 use PIN
                     00 requirement not used
                     (as in TS102221 - 9.5.2)

  AccAlways                = cardinal($00);
  NoAccess                 = cardinal($01);
  NeedPin1                 = cardinal($02);
  NeedPin2                 = cardinal($04);
  NeedPinU                 = cardinal($08);
  NeedAdm5                 = cardinal($10);
  NeedAdm6                 = cardinal($20);


80 01 01 | 90 00 | 80 01 1E | A0 18 A4 06 83 01 0A 95 01 08 A4 06 83 01 0B 95 01 08 A4 06 83 01 0D 95 01 08 |
80 01 01 | 90 00 | 80 01 1E | A0 10 A4 06 83 01 0A 95 01 08 A4 06 83 01 0B 95 01 08 |
80 01 01 | 90 00 | 80 01 06 | A0 20 A4 06 83 01 01 95 01 08 A4 06 83 01 0A 95 01 08 A4 06 83 01 0B 95 01 08 A4 06 83 01 0D 95 01 08 | 80 01 18 | A0 18 A4 06 83 01 0A 95 01 08 A4 06 83 01 0B 95 01 08 A4 06 83 01 0D 95 01 08 |
80 01 01 | 90 00 | 80 01 1E | A4 06 83 01 0A 95 01 08 |
80 01 1F | A0 10 A4 06 83 01 0A 95 01 08 A4 06 83 01 0B 95 01 08 |
80 01 07 | A0 18 A4 06 83 01 0A 95 01 08 A4 06 83 01 0B 95 01 08 A4 06 83 01 0D 95 01 08 |
FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF 90 00

  AccRuleRead              = $01;
  AccRuleUpdate            = $02;
  AccRuleActivate          = $10;
  AccRuleDeactivate        = $08;
  AccRuleIncrease          = $32;


var
  i : integer;
  ReadAccInfo,
  WriteAccInfo : boolean;
  Tlv          : TTlvContainer;
begin
  R := NoAccess;
  W := NoAccess;
  if Length(Rule) > 4 then
    begin
    Tlv := TTlvContainer.Create;
    try
      i := 0;
      SeparateTlvObjects(Rule,Tlv);
      while i < Tlv.Count do
        begin
        case Tlv[i].Tag of
          $80 : begin
                ReadAccInfo  := (OrdD(Tlv[i].Value,1) and AccRuleRead) > 0;
                WriteAccInfo := (OrdD(Tlv[i].Value,1) and AccRuleUpdate) > 0;
                Inc(i);
                if i < Tlv.Count then
                  begin
                  case Tlv[i].Tag of
                    $90 : begin
                          if ReadAccInfo  then R := AccAlways;
                          if WriteAccInfo then W := AccAlways;
                          end;
                    $97 : begin
                          if ReadAccInfo  then R := NoAccess;
                          if WriteAccInfo then W := NoAccess;
                          end;
                    $A0 : begin
                          if ReadAccInfo  then R := KeyrefUsequalToAccessBitmap(Tlv[i].Value,OrRelation);
                          if WriteAccInfo then W := KeyrefUsequalToAccessBitmap(Tlv[i].Value,OrRelation);
                          end;
                    $A4 : begin
                          if ReadAccInfo  then R := KeyrefUsequalToAccessBitmap(Tlv[i].Value,AccAlways);
                          if WriteAccInfo then W := KeyrefUsequalToAccessBitmap(Tlv[i].Value,AccAlways);
                          end;
                    end;
                  Inc(i);
                  end;
                end;
          end;
        end;
    finally
      Tlv.Free;
    end;
    end;
end;

procedure DecodeAccessRules(const RawRules: string; const RecLen: integer; var Rules: TAccessRules);
var
  i       : integer;
  NumRecs : integer;
  Rule    : string;
begin
  NumRecs := Length(RawRules) div RecLen;
  SetLength(Rules,NumRecs);
  for i := 0 to NumRecs - 1 do
    begin
    Rule := Copy(RawRules,i * RecLen,RecLen);
    DecodeExpandedAccessRule(Rule,Rules[i].ReadAccess,Rules[i].WriteAccess);
    end;
end;

}
////////////////////////////////////////////////////////////////////////////////
// object functions and procedures follow here

constructor TPCSCConnector.Create;
begin
  inherited Create;
  SystemInfo.dwOSVersionInfoSize := Sizeof(SystemInfo);
  GetVersionEx(SystemInfo);
  FReaderList   := TStringlist.Create;
  FContext      := 0;
  FCardHandle   := 0;
  FNumReaders   := 0;
  FUseReaderNum := -1;
  FConnected    := false;
  ActReaderState  := SCARD_STATE_UNAWARE;
  LastReaderState := SCARD_STATE_UNAWARE;
  ReaderOpen      := false;
  FChv1State      := csNotPresented;
  FChv2State      := csNotPresented;
  //FInitType       := itGSimFirst;
  ClearReaderAttributes;
  ClearCardAttributes;
  CheckIfPcscIsLoaded;
  NotifyHandle := AllocateHWnd(MessageWndProc);
end;

destructor TPCSCConnector.Destroy;
begin
  if PCSCLoaded then
    begin
    CloseAndDisconnect;
    SCardReleaseContext(FContext);
    end;
  FReaderList.Free;
  DeallocateHWnd(NotifyHandle);
  inherited Destroy;
end;

procedure TPCSCConnector.MessageWndProc(var Msg: TMessage);
begin
  if (Msg.Msg = WM_CARDSTATE) then
    begin
    if Msg.WParam <> SCARD_S_SUCCESS then
      if Assigned(FOnError) then FOnError(Self, esGetStatus, Msg.WParam);
    if ActReaderState <> LastReaderState then
      begin
      ProcessReaderState(LastReaderState, ActReaderState);
      end;
    end
    else Msg.Result := DefWindowProc(NotifyHandle, Msg.Msg, Msg.WParam, Msg.LParam);
end;

function TPCSCConnector.Init: boolean;
var
  RetVar         : cardinal;
  ReaderList     : string;
  ReaderListSize : integer;
  v              : array[0..MAXIMUM_SMARTCARD_READERS] of string;
  i              : integer;

begin
  Result      := false;
  FNumReaders := 0;
  if PCSCLoaded then
    begin
    CloseAndDisconnect;
    if CardIsValidContext(FContext) = SCARD_S_SUCCESS then SCardReleaseContext(FContext);
    RetVar := LongintToDword(SCardEstablishContext(SCARD_SCOPE_USER, nil, nil, @FContext));
    if RetVar = SCARD_S_SUCCESS then
      begin
      ReaderListSize := 0;
      RetVar := LongintToDword(SCardListReaders(FContext, nil, nil, ReaderListSize));
      if RetVar = SCARD_S_SUCCESS then
        begin
        SetLength(ReaderList, ReaderListSize);
        SCardListReaders(FContext, nil, Pointer(ReaderList), ReaderListSize);
        FReaderList.Clear;
        SortOutSubstrings(ReaderList,v,[#0]);
        for i := 0 to MAXIMUM_SMARTCARD_READERS do
          if v[i] <> '' then FReaderList.Add(v[i]);
        FNumReaders := FReaderList.Count;
        if FNumReaders > 0 then
          begin
          if Assigned(FOnReaderListChange) then FOnReaderListChange(Self);
          Result := true;
          end;
        end else if Assigned(FOnError) then FOnError(Self, esInit, RetVar);
      end else if Assigned(FOnError) then FOnError(Self, esInit, RetVar);
    end else if Assigned(FOnError) then FOnError(Self, esInit, SCARD_E_PCSC_NOTLOADED);
end;

function TPCSCConnector.Open: boolean;
var
  ThreadID    : LongWord;
begin
  CloseAndDisconnect;
  if (FUseReaderNum > NOREADERSELECTED) and
     (CardIsValidContext(FContext) = SCARD_S_SUCCESS) then
    begin
    ReaderOpen      := true;
    ActReaderState  := SCARD_STATE_UNAWARE;
    LastReaderState := SCARD_STATE_UNAWARE;
    BeginThread(nil, 0, CardWatcherThread, @FContext, 0, ThreadID);
    Result := true;
    end else Result := false;
end;

procedure TPCSCConnector.Close;
begin
  ReaderOpen := false;
  SCardCancel(FContext);
  if FConnected then Disconnect;
end;

function TPCSCConnector.Connect: boolean;
begin
  if FConnected then Disconnect;
  if FUseReaderNum > NOREADERSELECTED then
    if ConnectSelectedReader then FConnected := true
                             else FConnected := false;
  Result := FConnected;
end;

procedure TPCSCConnector.Disconnect;
begin
  if FConnected then
    begin
    SCardDisconnect(FCardHandle, SCARD_RESET_CARD);
    FConnected  := false;
    FCardHandle := 0;
    end;
end;

procedure TPCSCConnector.CloseAndDisconnect;
begin
  if FConnected then Disconnect;
  if ReaderOpen then Close;
end;

function TPCSCConnector.CardIsValidContext(Context: cardinal): cardinal;
begin
  if (SystemInfo.dwPlatformId = VER_PLATFORM_WIN32_NT) and
     (SystemInfo.dwMajorVersion >= 5) then
    begin
      Result := LongintToDword(SCardIsValidContext(Context));
    end
  else
    begin
      if Context <> 0 then Result := SCARD_S_SUCCESS
                      else Result := SCARD_E_INVALID_CONTEXT;
    end;
end;

function TPCSCConnector.ConnectSelectedReader: boolean;
var
  RetVar : cardinal;
begin
  RetVar := LongintToDword(SCardConnect(FContext,SelectedReader,
                                        SCARD_SHARE_EXCLUSIVE,SCARD_PROTOCOL_Tx,
                                        FCardHandle,@FAttrProtocol));
  case RetVar of
    SCARD_S_SUCCESS      : begin
                           //CardActiveAction;
                           Result := true;
                           end;
    SCARD_W_REMOVED_CARD : begin
                           Result := true;
                           end;
    else                   begin
                           Result := false;
                           if Assigned(FOnError) then FOnError(Self, esConnect, RetVar);
                           end;
    end;
end;

procedure TPCSCConnector.ProcessReaderState(const OldState,NewState: cardinal);
var
  CardInOld, CardInNew     : boolean;
  ReaderEmOld, ReaderEmNew : boolean;
  CardMuteOld, CardMuteNew : boolean;
  CardIgnore               : boolean;

begin
CardInOld   := (OldState and SCARD_STATE_PRESENT) > 0;
CardInNew   := (NewState and SCARD_STATE_PRESENT) > 0;
ReaderEmOld := (OldState and SCARD_STATE_EMPTY) > 0;
ReaderEmNew := (NewState and SCARD_STATE_EMPTY) > 0;
CardMuteOld := (OldState and SCARD_STATE_MUTE) > 0;
CardMuteNew := (NewState and SCARD_STATE_MUTE) > 0;
CardIgnore  := (NewState and SCARD_STATE_IGNORE) > 0;

LastReaderState := NewState;

//if Assigned(FOnLogMessage) then FOnLogMessage(Self,'ProcessReaderState','',IntToHex(OldState,8) + IntToHex(NewState,8));

if CardMuteNew     and
   not CardMuteOld then if Assigned(FOnCardInvalid) then FOnCardInvalid(Self);

if CardInNew       and
   not CardInOld   and
   not CardMuteNew and
   not CardIgnore  then CardInsertedAction;

if CardInOld     and
   not CardInNew then CardRemovedAction;

if ReaderEmNew     and
   not ReaderEmOld then begin
                        if Assigned(FOnReaderWaiting) then FOnReaderWaiting(Self);
                        end;

//if Assigned(FOnLogMessage) then FOnLogMessage(Self,'ProcessReaderState','End',IntToHex(LastReaderState,8));
end;

procedure TPCSCConnector.CardInsertedAction;
begin
  if Assigned(FOnCardInserted) then FOnCardInserted(Self);
  if FConnected then CardActiveAction;
end;

procedure TPCSCConnector.CardActiveAction;
begin
  GetReaderAttributes;
  if FAttrProtocol <> SCARD_PROTOCOL_UNK then
    begin
    GetCardAttributes;
    if Assigned(FOnCardActive) then FOnCardActive(Self);
    end;
end;

procedure TPCSCConnector.CardRemovedAction;
begin
  ClearReaderAttributes;
  ClearCardAttributes;
  if Assigned(FOnCardRemoved) then FOnCardRemoved(Self);
  Disconnect;
end;

procedure TPCSCConnector.SetReaderNum(Value: Integer);
begin
  if Value <> FUseReaderNum then
    begin
    CloseAndDisconnect;
    if Value < FReaderList.Count then
      begin
      SelectedReader := PChar(FReaderList[Value]);
      FUseReaderNum   := Value;
      end else
      begin
      SelectedReader := '';
      FUseReaderNum   := -1;
      end;
    end;
end;

function TPCSCConnector.IsReaderOpen: boolean;
begin
  Result := ReaderOpen;
end;

function TPCSCConnector.GetReaderState: cardinal;
begin
  Result := ActReaderState;
end;

procedure TPCSCConnector.GetReaderAttributes;
var
  RetVar : cardinal;
  ABuf   : string;
  AIBuf  : integer;
  ALen   : cardinal;
begin
  ABuf := StringOfChar(#0, 127);
  ALen := Length(ABuf);
  RetVar := LongintToDword(SCardGetAttrib(FCardHandle, SCARD_ATTR_ATR_STRING, Pointer(ABuf), @ALen));
  if RetVar = SCARD_S_SUCCESS then FAttrCardATR := Copy(ABuf, 1, ALen)
                              else FAttrCardATR := '';

  ALen := Length(ABuf);
  RetVar := SCardGetAttrib(FCardHandle, SCARD_ATTR_VENDOR_NAME, Pointer(ABuf), @ALen);
  if RetVar = SCARD_S_SUCCESS then FAttrVendorName := Copy(ABuf, 1, ALen)
                              else FAttrVendorName := '';

  ALen := Length(ABuf);
  RetVar := LongintToDword(SCardGetAttrib(FCardHandle, SCARD_ATTR_VENDOR_IFD_SERIAL_NO, Pointer(ABuf), @ALen));
  if RetVar = SCARD_S_SUCCESS then FAttrVendorSerial := Copy(ABuf, 1, ALen)
                              else FAttrVendorSerial := '';

  ALen := SizeOf(AIBuf);
  RetVar := LongintToDword(SCardGetAttrib(FCardHandle, SCARD_ATTR_CURRENT_PROTOCOL_TYPE, @AIBuf, @ALen));
  if RetVar = SCARD_S_SUCCESS then FAttrProtocol := AIBuf
                              else FAttrProtocol := 0;

  ALen := SizeOf(AIBuf);
  AIBuf := 0;
  RetVar := LongintToDword(SCardGetAttrib(FCardHandle, SCARD_ATTR_ICC_TYPE_PER_ATR, @AIBuf, @ALen));
  if RetVar = SCARD_S_SUCCESS then begin
                                   case AIBuf of
                                     1  : FAttrICCType := 'ISO7816A';
                                     2  : FAttrICCType := 'ISO7816S';
                                     else FAttrICCType := 'UNKNOWN';
                                     end;
                                   end
                              else FAttrICCType := '';
  if FDebugging then
    if Assigned(FOnLogMessage) then
      FOnLogMessage(Self,'CardAttrib','ATR',BinToHexExt(FAttrCardATR));

end;

procedure TPCSCConnector.ClearReaderAttributes;
begin
  FAttrCardATR      := '';
  FAttrVendorName   := '';
  FAttrVendorSerial := '';
  FAttrProtocol     := 0;
  FAttrICCType      := '';
end;

procedure TPCSCConnector.GetCardAttributes;
var
  Status        : string;
  SimState      : word;

  procedure InitGsmCard;
  begin
    SimState := CardCommandSelect(WordToBytes(DFgsm),Status);
    if SimState = SIM_STATE_OK then
      begin
      //CardCommandStatus(Status);
      case (OrdD(Status, 14) and $70) of
        $00 : FCardVoltage := vc50Volt;
        $10 : FCardVoltage := vc30Volt;
        $30 : FCardVoltage := vc18Volt;
        $70 : FCardVoltage := vcOtherClass;
        else  FCardVoltage := vcUnknown;
        end;
      end;
  end;

  procedure InitTETRACard;
  begin
    SimState := CardCommandSelect(WordToBytes(DFtetra),Status);
    if SimState = SIM_STATE_OK then
      begin
      case (OrdD(Status, 14) and $70) of
        $00 : FCardVoltage := vc50Volt;
        $10 : FCardVoltage := vc30Volt;
        $30 : FCardVoltage := vc18Volt;
        $70 : FCardVoltage := vcOtherClass;
        else  FCardVoltage := vcUnknown;
        end;
      end;
  end;

  procedure AnalyzeAtr;
  var
    HistStart : integer;
    IfChar    : integer;
  begin
    IfChar    := 2;
    HistStart := 3;
    repeat
      if (OrdD(FAttrCardATR,IfChar) and $10) > 0 then Inc(HistStart);
      if (OrdD(FAttrCardATR,IfChar) and $20) > 0 then Inc(HistStart);
      if (OrdD(FAttrCardATR,IfChar) and $40) > 0 then Inc(HistStart);
      if (OrdD(FAttrCardATR,IfChar) and $80) > 0 then
        begin
        Inc(HistStart);
        IfChar := HistStart - 1;
        end
      else
        begin
        IfChar := 0;
        end;
    until IfChar = 0;
    if (OrdD(FAttrCardATR,HistStart) = $80) and
       (OrdD(FAttrCardATR,HistStart + 1) = $31) and
       (OrdD(FAttrCardATR,HistStart + 3) = $73) then ChangeCardClass(ccUicc)
                                                else ChangeCardClass(ccUnknown);
  end;

begin
  //AnalyzeAtr;

  ChangeCardClass(FExpectCardClass);
  if FCardClass = ccGsmSim then InitGsmCard;
  if FCardClass = ccTSim   then InitTETRACard;

  if CardCommandSelect(WordToBytes(EFdir),Status) = SIM_STATE_OK then
    begin
    case FCardClass of
      ccGsmSim  : ChangeCardClass(ccJanus);
      ccUnknown : ChangeCardClass(ccOtherApp);
      end;
    end;
  CardSelectMF(Status);
end;

procedure TPCSCConnector.ClearCardAttributes;
begin
  FCurrentFile := '';
  FLastStatus  := '';
  FCardVoltage := vcUnknown;
  FChv1State   := csNotPresented;
  FChv2State   := csNotPresented;
  CardClass    := ccUnknown;
  ClearFileInfo(FLastFileInfo);
end;

procedure TPCSCConnector.ChangeCardClass(const NewClass: TCardClass);
begin
  case NewClass of
    ccGsmSim,
//    ccTSim,
    ccJanus  : begin
               CmdGetStatus    := GsmCmdGetStatus;
               CmdGetResponse  := GsmCmdGetResponse;
               CmdSelectFile   := GsmCmdSelectFile;
               CmdReadBinary   := GsmCmdReadBinary;
               CmdReadRecord   := GsmCmdReadRecord;
               CmdReadCurrRec  := GsmCmdReadCurrRecord;
               CmdReadNextRec  := GsmCmdReadNextRecord;
               CmdUpdateBinary := GsmCmdUpdateBinary;
               CmdUpdateRecord := GsmCmdUpdateRecord;
               CmdVerifyChv    := GsmCmdVerifyChv;
               CmdChangeChv    := GsmCmdChangeChv;
               CmdUnblockChv   := GsmCmdUnblockChv;
               end;
    ccUicc,
    ccTSim,
    ccUSim :   begin
               CmdGetStatus    := UsimCmdGetStatus;
               CmdGetResponse  := UsimCmdGetResponse;
               CmdSelectFile   := UsimCmdSelectFile;
               CmdReadBinary   := UsimCmdReadBinary;
               CmdReadRecord   := UsimCmdReadRecord;
               CmdReadCurrRec  := UsimCmdReadCurrRecord;
               CmdReadNextRec  := UsimCmdReadNextRecord;
               CmdUpdateBinary := UsimCmdUpdateBinary;
               CmdUpdateRecord := UsimCmdUpdateRecord;
               CmdVerifyChv    := UsimCmdVerifyChv;
               CmdChangeChv    := UsimCmdChangeChv;
               CmdUnblockChv   := UsimCmdUnblockChv;
               end;
    else       begin
               CmdGetStatus    := GsmCmdGetStatus;
               CmdGetResponse  := GsmCmdGetResponse;
               CmdSelectFile   := GsmCmdSelectFile;
               CmdReadBinary   := GsmCmdReadBinary;
               CmdReadRecord   := GsmCmdReadRecord;
               CmdReadCurrRec  := GsmCmdReadCurrRecord;
               CmdReadNextRec  := GsmCmdReadNextRecord;
               CmdUpdateBinary := GsmCmdUpdateBinary;
               CmdUpdateRecord := GsmCmdUpdateRecord;
               CmdVerifyChv    := GsmCmdVerifyChv;
               CmdChangeChv    := GsmCmdChangeChv;
               CmdUnblockChv   := GsmCmdUnblockChv;
               end;
    end;
  if FDebugging then
    if Assigned(FOnLogMessage) then
      FOnLogMessage(Self,'CardClass','Change',DecodeCardClass(FCardClass) + ' -> ' + DecodeCardClass(NewClass));
  FCardClass := NewClass;
end;

procedure TPCSCConnector.CheckIfPcscIsLoaded;
begin
  PCSCLoaded := false;
  try
    SCardDisconnect(0,0);
    PCSCLoaded := true;
  finally
  end;
end;

////////////////////////////////////////////////////////////////////////////////
// reading from the card, layer 1: direct card access

function TPCSCConnector.SendPduToCard(const APdu: string): string;
var
  RetVar : cardinal;
  SBuf   : string;
  SLen   : cardinal;
  RBuf   : string;
  RLen   : cardinal;
  Ppci   : Pointer;
begin
  SBuf := APdu;
  RBuf := StringOfChar(#0,MAXAPDULENGTH);
  if Length(SBuf) <= MAXAPDULENGTH then
    begin
    case FAttrProtocol of
      SCARD_PROTOCOL_T0 : Ppci := @SCARD_PCI_T0;
      SCARD_PROTOCOL_T1 : Ppci := @SCARD_PCI_T1;
      else                Ppci := nil;
      end;
    SLen := Length(APdu);
    RLen := Length(RBuf);

    if FDebugging then
      if Assigned(FOnLogMessage) then
        FOnLogMessage(Self,'SendPduToCard','StartSend',DecodeCommandFromPdu(APdu) + BinToHex(APdu));

    RetVar := LongintToDword(SCardTransmit(FCardHandle, Ppci, Pointer(SBuf), SLen, nil, Pointer(RBuf), @RLen));

    if RetVar = SCARD_S_SUCCESS then
      begin
      Result := Copy(RBuf,1,RLen);
      if FDebugging then
        if Assigned(FOnLogMessage) then
          begin
          if Length(Result) = 2 then FOnLogMessage(Self,'SendPduToCard','Result',DecodeResponseCode(BytesToCardinal(Result)))
                                else FOnLogMessage(Self,'SendPduToCard','Result',BinToHexExt(Result));
          end;
      end
    else
      begin
      Result := '';
      if Assigned(FOnError) then FOnError(Self, esTransmit, RetVar);
      end;
  end;
end;

function TPCSCConnector.SendCommandToCard(const Command: string; var Data: string; var Status: word): boolean;
var
  Answer  : string;
  AnswerL : integer;
  sw1,sw2 : byte;
begin
  Answer := SendPduToCard(Command + Data);
  AnswerL := Length(Answer);
  if AnswerL >= 2 then
    begin
    Data := Copy(Answer, 1, AnswerL - 2);
    sw1  := Ord(Answer[AnswerL - 1]);
    sw2  := Ord(Answer[AnswerL]);
    Status := (sw1 shl 8) or sw2;

{    if FDebugging then
      if Assigned(FOnLogMessage) then FOnLogMessage(Self,'SendCommandToCard','Status',DecodeResponseCode(Status));
}
    if sw1 in [SIM_STATE_HAVEDATA,SIM_STATE_DOGETRESPONSE] then
      begin
      Data := Chr(sw2);
      if not SendCommandToCard(CmdGetResponse, Data, Status) then
        begin
        Data := '';
        Result := false;
        end
      else Result := true;
      end
    else Result := true;
    end
  else
    begin
    Data := '';
    Status := SIM_STATE_TECHPROBLEMS;
    Result := false;
    end;
end;

////////////////////////////////////////////////////////////////////////////////
// reading from the card, layer 2: mapping of the commands from TS GSM 11.11 / 3GPP 31.101

{
Command              Input    Output        Command

SELECT         GSM   FileID   Status Info   A0 A4 00         00           02
SELECT        USIM   FileID   FCP Template  00 A4 00         04           02
SELECT        USIM   AID      FCP Template  00 A4 04         04           lgth

STATUS         GSM   -        Status Info   A0 F2 00         00           lgth
STATUS        USIM   -        Status Info   80 F2 00         00           lgth

READ BINARY    GSM   Offset                 A0 B0 OffsetHigh OffsetLow    lgth
READ BINARY   USIM   Offset                 00 B0 OffsetHigh OffsetLow    lgth

UPDATE BINARY  GSM   Offset                 A0 D6 OffsetHigh OffsetLow    lgth
UPDATE BINARY USIM   Offset                 00 D6 OffsetHigh OffsetLow    lgth

READ RECORD    GSM                          A0 B2 RecordNum  Mode         lgth
READ RECORD   USIM                          00 B2 RecordNum  Mode         lgth

UPDATE RECORD  GSM                          A0 DC XX         YY           lgth
UPDATE RECORD USIM                          00 DC XX         YY           lgth
                                                  00         02:next      lgth
                                                  00         03:previous  lgth
                                                  RecordNum  04:absolute  lgth

VERIFY CHV     GSM   Chv                    A0 20 00         CHV No.   08
VERIFY CHV    USIM   Chv                    00 20 00         CHV No.   08

DISABLE CHV    GSM   Chv                    A0 26 00 01 08
DISABLE CHV   USIM   Chv                    00 26 00 01 08

ENABLE CHV     GSM   Chv                    A0 28 00 01 08
ENABLE CHV    USIM   Chv                    00 28 00 01 08

UNBLOCK CHV    GSM   Chv                    A0 2C 00         CHV No.   10
UNBLOCK CHV   USIM   Chv                    00 2C 00         CHV No.   10

CHANGE CHV     GSM   ChvOld, ChvNew         00 24 00         CHV No.   10

CHANGE PIN 0X '24'
}

function TPCSCConnector.CardCommandStatus(var Status: string): word;
begin
  SendCommandToCard(CmdGetStatus, Status, Result);
  if Result = SIM_STATE_OK then
    begin
    FCurrentFile := FLastFileInfo.FileId;
    FLastStatus  := Status;
    end else
    begin
    FLastStatus  := '';
    ClearFileInfo(FLastFileInfo);
    end;
end;

function TPCSCConnector.CardCommandSelect(const FileID: string; var Status: string): word;
var
  StatusWord        : word;
  AccessRuleContent : string;
begin
  Status := FileID;

  if (FileId = #$3F#$00) then
    begin
    case FCardClass of
      ccGsmSim : SendCommandToCard(CmdSelectFile,Status,StatusWord);
      else       SendCommandToCard(UsimCmdSelectMF,Status,StatusWord);
      end;
    end
  else SendCommandToCard(CmdSelectFile,Status,StatusWord);

  DecodeFileStatus(Status,FLastFileInfo);
  if StatusWord = SIM_STATE_OK then
    begin
    if Assigned(FOnProgress) then FOnProgress(Self,FileID,0);
    if FLastFileInfo.AccessRules <> '' then
      begin
      if FLastFileInfo.AccessRules <> FAccessRuleFile then // Caution! This may not work under all circumstances
        begin
        if CardReadFile(FLastFileInfo.AccessRules,Status,AccessRuleContent) = SIM_STATE_OK then
          begin
          //DecodeAccessRules(AccessRuleContent,FLastFileInfo.RecordLen,FAccessRules);
          end
        else
          begin
          SetLength(FAccessRules,0);
          end ;
        Status := FileID;
        SendCommandToCard(CmdSelectFile,Status,StatusWord);
        DecodeFileStatus(Status,FLastFileInfo);
        end;
      end;
    // this may need a change for USIM/TSIM because the AppDirectory does not in all cases have a FileID
    FCurrentFile := FLastFileInfo.FileId;
    FLastStatus  := Status;
    end
  else
    begin
    FCurrentFile := '';
    FLastStatus  := '';
    ClearFileInfo(FLastFileInfo);
    end;
  Result := StatusWord;
end;

function TPCSCConnector.CardCommandReadBinary(const Offset, Length: integer; var Data: string): word;
var
  Command  : string;
begin
  Command := CmdReadBinary + Chr((Offset div 256) mod 128) + Chr(Offset mod 256) + Chr(Length mod 256);
  SendCommandToCard(Command, Data, Result);
end;

function TPCSCConnector.CardCommandUpdateBinary(const Offset: integer; var Data: string): word;
var
  Command  : string;
begin
  Command := CmdUpdateBinary + Chr((Offset div 256) mod 128) + Chr(Offset mod 256) + Chr(Length(Data) mod 256);
  SendCommandToCard(Command, Data, Result);
end;

function TPCSCConnector.CardCommandReadRecord(const RecordNum, Length: integer; var Data: string): word;
var
  Command  : string;
begin
  Command := CmdReadRecord + Chr(RecordNum mod 256) + Chr($04) + Chr(Length mod 256);
  SendCommandToCard(Command, Data, Result);
end;

function TPCSCConnector.CardCommandUpdateRecord(const RecordNum: integer; var Data: string): word;
var
  Command  : string;
begin
  Command := CmdUpdateRecord + Chr(RecordNum mod 256) + Chr($04) + Chr(Length(Data) mod 256);
  SendCommandToCard(Command, Data, Result);
end;

function TPCSCConnector.CardCommandReadCurrentRecord(const Length: integer; var Data: string): word;
var
  Command  : string;
begin
  Command := CmdReadCurrRec + Chr(Length mod 256);
  SendCommandToCard(Command, Data, Result);
end;

function TPCSCConnector.CardCommandReadNextRecord(const Length: integer; var Data: string): word;
var
  Command  : string;
begin
  Command := CmdReadNextRec + Chr(Length mod 256);
  SendCommandToCard(Command, Data, Result);
end;

function TPCSCConnector.CardCommandVerifyChv(const ChvType: TNeededPIN; const ChvValue: string): word;
var
  Command   : string;
  DummyData : string;
begin
  case ChvType of
    npPin1 : Command := CmdVerifyChv + #$01 + #$08 + ChvValue;
    npPin2 : Command := CmdVerifyChv + #$02 + #$08 + ChvValue;
    else     Command := '';
  end;
  SendCommandToCard(Command, DummyData, Result);
end;

function TPCSCConnector.CardCommandChangeChv(const ChvType: TNeededPIN; const OldValue,NewValue: string): word;
var
  Command   : string;
  DummyData : string;
begin
  case ChvType of
    npPin1 : Command := CmdChangeChv + #$01 + #$10 + OldValue + NewValue;
    npPin2 : Command := CmdChangeChv + #$02 + #$10 + OldValue + NewValue;
    else     Command := '';
  end;
  SendCommandToCard(Command, DummyData, Result);
end;

function TPCSCConnector.CardCommandUnblockChv(const ChvType: TNeededPIN; const UnblockChv,NewChv: string): word;
var
  Command   : string;
  DummyData : string;
begin
  case ChvType of
    npPUK1 : if CardClass in [ccUicc,ccUsim] // TS GSM 11.11, 9.2.13
               then Command := CmdUnblockChv + #$01 + #$10 + UnblockChv + NewChv
               else Command := CmdUnblockChv + #$00 + #$10 + UnblockChv + NewChv;
    npPUK2 : Command := CmdUnblockChv + #$02 + #$10 + UnblockChv + NewChv;
    else     Command := '';
  end;
  SendCommandToCard(Command, DummyData, Result);
end;

////////////////////////////////////////////////////////////////////////////////
// reading from the card, layer 3: more abstract things like reading a file

function TPCSCConnector.CardCheckAccess(const Needed: TAccessCondition): boolean;
begin
  case Needed of
    acUnknown : Result := true;
    acAlways  : Result := true;
    acChv1    : case FChv1State of
                  csChvOk       : Result := true;
                  csChvDisabled : Result := true;
                  csChvRefused  : Result := false;
                  csBlocked     : Result := false;
                  else begin
                       if Assigned(FOnCHVNeeded) then FOnCHVNeeded(Self,npPIN1);
                       if FChv1State = csChvOk then Result := true
                                               else Result := false;
                       end;
                  end;
    acChv2    : case FChv2State of
                  csChvOk       : Result := true;
                  csChvRefused  : Result := false;
                  csBlocked     : Result := false;
                  else begin
                       if Assigned(FOnCHVNeeded) then FOnCHVNeeded(Self,npPIN2);
                       if FChv2State = csChvOk then Result := true
                                               else Result := false;
                       end;
                  end;
    acRfu     : Result := false;
    acAdm     : Result := false;
    acNever   : Result := false;
    else        Result := false;
    end;
end;

function TPCSCConnector.CardReadTransparentFile(const FileID: string; var Status,Data: string): word;
begin
  if FileID = FCurrentFile then Status := FLastStatus
                           else CardCommandSelect(FileID, Status);
  if FileID = FCurrentFile then
    begin
    if CardCheckAccess(FLastFileInfo.ReadCond) then
      begin
      if FLastFileInfo.FileStruct = fsTransparent then
        begin
        Result := CardCommandReadBinary(0,FLastFileInfo.FileSize,Data);
        end else Result := SIM_STATE_WRONGSTRUCTURE;
      end else Result := SIM_STATE_ACCESSDENIED;
    end else Result := SIM_STATE_FILENOTFOUND;
end;

function TPCSCConnector.CardReadRecordsTopDown(const FileID: string; var Status,Data: string): word;
var
  i        : integer;
  s        : string;
begin
  Result := SIM_STATE_TECHPROBLEMS; //because of a enerving compiler warning
  if FileID = FCurrentFile then Status := FLastStatus
                           else CardCommandSelect(FileID, Status);
  if FileID = FCurrentFile then
    begin
    if CardCheckAccess(FLastFileInfo.ReadCond) then
      begin
      if FLastFileInfo.FileStruct in [fsLinearFixed,fsCyclic] then
        begin
        if FLastFileInfo.NumRecords = 0 then Result := SIM_STATE_P1P2INCORRECT;
        for i := 1 to FLastFileInfo.NumRecords do
          begin
          s := '';
          if Assigned(FOnProgress) then FOnProgress(Self,FileID,i);
          Result := CardCommandReadRecord(i,FLastFileInfo.RecordLen,s);
          if Result = SIM_STATE_OK then Data := Data + s
                                   else Data := Data + StringOfChar(#$FF,FLastFileInfo.RecordLen);
          end;
        end else Result := SIM_STATE_WRONGSTRUCTURE;
      end else Result := SIM_STATE_ACCESSDENIED;
    end else Result := SIM_STATE_FILENOTFOUND;
end;

function TPCSCConnector.CardReadRecordsCyclic(const FileID: string; var Status,Data: string): word;
var
  i        : integer;
  s        : string;
begin
  //Result := SIM_STATE_TECHPROBLEMS; //because of a enerving compiler warning
  if FileID = FCurrentFile then Status := FLastStatus
                           else CardCommandSelect(FileID, Status);
  if FileID = FCurrentFile then
    begin
    if CardCheckAccess(FLastFileInfo.ReadCond) then
      begin
      if FLastFileInfo.FileStruct = fsCyclic then
        begin
        if FLastFileInfo.NumRecords = 0 then
          Result := SIM_STATE_P1P2INCORRECT
        else
          begin
          s := '';
          Result := CardCommandReadCurrentRecord(FLastFileInfo.RecordLen,s);
          if Result = SIM_STATE_OK then Data := Data + s
                                   else Data := Data + StringOfChar(#$FF,FLastFileInfo.RecordLen);
          end;
        for i := 1 to FLastFileInfo.NumRecords - 1 do
          begin
          s := '';
          if Assigned(FOnProgress) then FOnProgress(Self,FileID,i);
          Result := CardCommandReadNextRecord(FLastFileInfo.RecordLen,s);
          if Result = SIM_STATE_OK then Data := Data + s
                                   else Data := Data + StringOfChar(#$FF,FLastFileInfo.RecordLen);
          end;
        end else Result := SIM_STATE_WRONGSTRUCTURE;
      end else Result := SIM_STATE_ACCESSDENIED;
    end else Result := SIM_STATE_FILENOTFOUND;
end;

function TPCSCConnector.CardUpdateRecord(const FileID: string; const RecordNum: integer; var Data: string): word;
var
  Status   : string;
begin
  if FileID = FCurrentFile then Status := FLastStatus
                           else CardCommandSelect(FileID, Status);
  if FileID = FCurrentFile then
    begin
    if CardCheckAccess(FLastFileInfo.WriteCond) then
      begin
      if FLastFileInfo.FileStruct in [fsLinearFixed,fsCyclic] then
        begin
        if Length(Data) < FLastFileInfo.RecordLen
          then Data := Data + StringOfChar(#$FF,FLastFileInfo.RecordLen - Length(Data))
          else if Length(Data) > FLastFileInfo.RecordLen then Data := Copy(Data,1,FLastFileInfo.RecordLen);
        Result := CardCommandUpdateRecord(RecordNum,Data);
        end else Result := SIM_STATE_WRONGSTRUCTURE;
      end else Result := SIM_STATE_ACCESSDENIED;
    end else Result := SIM_STATE_FILENOTFOUND;
end;

function TPCSCConnector.CardUpdateBinary(const FileID: string; const Offset: integer; var Data: string): word;
var
  Status   : string;
begin
  if FileID = FCurrentFile then Status := FLastStatus
                           else CardCommandSelect(FileID, Status);
  if FileID = FCurrentFile then
    begin
    if CardCheckAccess(FLastFileInfo.WriteCond) then
      begin
      if FLastFileInfo.FileStruct = fsTransparent then
        begin
        if (Offset + Length(Data)) > FLastFileInfo.FileSize
          then Data := Copy(Data,1,FLastFileInfo.FileSize - Offset);
        Result := CardCommandUpdateBinary(Offset,Data);
        end else Result := SIM_STATE_WRONGSTRUCTURE;
      end else Result := SIM_STATE_ACCESSDENIED;
    end else Result := SIM_STATE_FILENOTFOUND;
end;

function TPCSCConnector.CardReadFile(const FileID: string; var Status,Data: string): word;
begin
  if FileID = FCurrentFile then Status := FLastStatus
                           else CardCommandSelect(FileID, Status);
  if FileID = FCurrentFile then
    begin
    case FLastFileInfo.FileStruct of
      fsTransparent : Result := CardReadTransparentFile(FileID,Status,Data);
      fsLinearFixed : Result := CardReadRecordsTopDown(FileID,Status,Data);
      fsCyclic      : Result := CardReadRecordsCyclic(FileID,Status,Data);
      else            begin
                      Result := SIM_STATE_WRONGSTRUCTURE;
                      Data   := '';
                      end;
      end;
    end else
    begin
    Result := SIM_STATE_FILENOTFOUND;
    Data   := '';
    end;
end;

function TPCSCConnector.CardSelectMF(var Status: string): word;
begin
  Result := CardCommandSelect(WordToBytes(DFmaster),Status);
end;

function TPCSCConnector.CardChangeDir(const FileID: string; var Status: string): word;
begin
  Result := CardCommandSelect(FileID, Status);
end;

function TPCSCConnector.CardChangeDir(const FileID: word; var Status: string): word;
begin
  Result := CardCommandSelect(WordToBytes(FileID), Status);
end;

function TPCSCConnector.CardVerifyChv(const ChvType: TNeededPIN; const ChvValue: string): word;
var
  TempChv : string;
  TempRes : word;
begin
  case ChvType of
    npPIN1,
    npPIN2 : begin
             TempChv := FilterNumbers(ChvValue);
             TempChv := TempChv + StringOfChar(#$FF,8 - Length(TempChv));
             TempRes := CardCommandVerifyChv(ChvType, TempChv);
             case TempRes of
               SIM_STATE_OK         : begin
                                      if ChvType = npPIN1 then FChv1State := csChvOk
                                                          else FChv2State := csChvOk;
                                      end;
               SIM_STATE_CHVBLOCKED : begin
                                      if ChvType = npPIN1 then FChv1State := csBlocked
                                                          else FChv2State := csBlocked;
                                      end;
               else                   begin
                                      if ChvType = npPIN1 then FChv1State := csChvNotOk
                                                          else FChv2State := csChvNotOk;
                                      end;
               end;
             Result := TempRes;
             end;
    else     Result := SIM_STATE_P1P2INCORRECT;
    end;
end;

function TPCSCConnector.CardChangeChv(const ChvType: TNeededPIN; const OldValue,NewValue: string): word;
var
  TempOldChv : string;
  TempNewChv : string;
  TempRes    : word;
begin
  case ChvType of
    npPIN1,
    npPIN2 : begin
             TempOldChv := FilterNumbers(OldValue);
             TempNewChv := FilterNumbers(NewValue);
             TempOldChv := TempOldChv + StringOfChar(#$FF,8 - Length(TempOldChv));
             TempNewChv := TempNewChv + StringOfChar(#$FF,8 - Length(TempNewChv));
             TempRes    := CardCommandChangeChv(ChvType,TempOldChv,TempNewChv);
             case TempRes of
               SIM_STATE_OK         : begin
                                      if ChvType = npPIN1 then FChv1State := csChvOk
                                                          else FChv2State := csChvOk;
                                      end;
               SIM_STATE_CHVBLOCKED : begin
                                      if ChvType = npPIN1 then FChv1State := csBlocked
                                                          else FChv2State := csBlocked;
                                      end;
               else                   begin
                                      if ChvType = npPIN1 then FChv1State := csChvNotOk
                                                          else FChv2State := csChvNotOk;
                                      end;
               end;
             Result := TempRes;
             end;
    else     Result := SIM_STATE_P1P2INCORRECT;
    end;
end;

function TPCSCConnector.CardUnblockChv(const ChvType: TNeededPIN; const UnblockChv,NewValue: string): word;
var
  TempUnbChv : string;
  TempNewChv : string;
  TempRes    : word;
begin
  case ChvType of
    npPUK1,
    npPUK2 : begin
             TempUnbChv := FilterNumbers(UnblockChv);
             TempNewChv := FilterNumbers(NewValue);
             TempUnbChv := TempUnbChv + StringOfChar(#$FF,8 - Length(TempUnbChv));
             TempNewChv := TempNewChv + StringOfChar(#$FF,8 - Length(TempNewChv));
             TempRes    := CardCommandUnblockChv(ChvType,TempUnbChv,TempNewChv);
             case TempRes of
               SIM_STATE_OK         : begin
                                      if ChvType = npPUK1 then FChv1State := csChvOk
                                                          else FChv2State := csChvOk;
                                      end;
               SIM_STATE_ACCESSDENIED,
               SIM_STATE_CHVBLOCKED : begin
                                      if ChvType = npPUK1 then FChv1State := csBlocked
                                                          else FChv2State := csBlocked;
                                      end;
               else                   begin
                                      if ChvType = npPUK1 then FChv1State := csChvNotOk
                                                          else FChv2State := csChvNotOk;
                                      end;
               end;
             Result := TempRes;
             end;
    else     Result := SIM_STATE_P1P2INCORRECT;
    end;
end;

////////////////////////////////////////////////////////////////////////////////
// reading from the card, layers above 3: things like decoding routines etc.

procedure TPCSCConnector.ClearFileInfo(var FileInfo: TFileInfo);
begin
  FileInfo.FileType   := ftUnknown;
  FileInfo.FileId     := '';
  FileInfo.AccessRules := '';
  // for a EFile
  FileInfo.FileStruct := fsNone;
  FileInfo.FileSize   := 0;
  FileInfo.RecordLen  := 0;
  FileInfo.NumRecords := 0;
  FileInfo.ReadCond   := acUnknown;
  FileInfo.WriteCond  := acUnknown;
  // for a DFile
  FileInfo.AppId       := '';
  FileInfo.BytesFree   := 0;
  FileInfo.NumDFiles   := 0;
  FileInfo.NumEFiles   := 0;
  FileInfo.NumCodes    := 0;
  FileInfo.Chv1Enable  := true;
  FileInfo.Chv1Tries   := 0;
  FileInfo.Chv2Tries   := 0;
  FileInfo.Chv1UTries  := 0;
  FileInfo.Chv2UTries  := 0;
  FileInfo.ClockStop   := csNotAllowed;
end;

procedure TPCSCConnector.DecodeFileStatus(const Status: string; var FileInfo: TFileInfo);

  function GetAccCondFromAccessRule(const AccBitmap: TAccessBitmap): TAccessCondition;
  begin
    if AccBitmap = AccAlways then Result := acAlways
    else if (AccBitmap and NoAccess) > 0 then Result := acNever
    else if (AccBitmap and NeedPin1) > 0 then Result := acChv1
    else if (AccBitmap and NeedPin2) > 0 then Result := acChv2
    else if (AccBitmap and NeedPinU) > 0 then Result := acChvU
    else if (AccBitmap and NeedAdm5) > 0 then Result := acAdm
    else if (AccBitmap and NeedAdm6) > 0 then Result := acAdm
    else Result := acUnknown;
  end;

  procedure DecodeFileStatusGsmEFile(const Status: string; var FileInfo: TFileInfo);
  begin
    if Length(Status) >= 15 then
      begin
      case OrdD(Status,14) of
        0 :  FileInfo.FileStruct := fsTransparent;
        1 :  FileInfo.FileStruct := fsLinearFixed;
        3 :  FileInfo.FileStruct := fsCyclic;
        else FileInfo.FileStruct := fsUnknown;
        end;
      FileInfo.FileSize  := (OrdD(Status,3) shl 8) + OrdD(Status,4);
      FileInfo.RecordLen := OrdD(Status,15);
      if FileInfo.FileStruct = fsTransparent
        then FileInfo.NumRecords := 0
        else FileInfo.NumRecords := FileInfo.FileSize div FileInfo.RecordLen;
      case (OrdD(Status,9) shr 4) of
         0     : FileInfo.ReadCond := acAlways;
         1     : FileInfo.ReadCond := acChv1;
         2     : FileInfo.ReadCond := acChv2;
         3     : FileInfo.ReadCond := acRfu;
         4..14 : FileInfo.ReadCond := acAdm;
        15     : FileInfo.ReadCond := acNever;
        end;
      case (OrdD(Status,9) and $0F) of
         0     : FileInfo.WriteCond := acAlways;
         1     : FileInfo.WriteCond := acChv1;
         2     : FileInfo.WriteCond := acChv2;
         3     : FileInfo.WriteCond := acRfu;
         4..14 : FileInfo.WriteCond := acAdm;
        15     : FileInfo.WriteCond := acNever;
        end;
      end;
  end;

  procedure DecodeFileStatusGsmDFile(const Status: string; var FileInfo: TFileInfo);
  begin
    if Length(Status) >= 22 then
      begin
      FileInfo.BytesFree  := (OrdD(Status,3) shl 8) + OrdD(Status,4);
      FileInfo.NumDFiles  := OrdD(Status,15);
      FileInfo.NumEFiles  := OrdD(Status,16);
      FileInfo.NumCodes   := OrdD(Status,17);
      FileInfo.Chv1Enable := (OrdD(Status,14) and $80) <> 0;
      FileInfo.Chv1Tries  := OrdD(Status,19) and $0F;
      FileInfo.Chv2Tries  := OrdD(Status,21) and $0F;
      FileInfo.Chv1UTries := OrdD(Status,20) and $0F;
      FileInfo.Chv2UTries := OrdD(Status,22) and $0F;
      case OrdD(Status,14) and $0D of
        $00 : FileInfo.ClockStop := csNotAllowed;
        $04 : FileInfo.ClockStop := csOnlyHigh;
        $08 : FileInfo.ClockStop := csOnlyLow;
        $01 : FileInfo.ClockStop := csNoPref;
        $05 : FileInfo.ClockStop := csHighPref;
        $09 : FileInfo.ClockStop := csLowPref;
        else  FileInfo.ClockStop := csNotAllowed;
        end;
      end;
  end;

  procedure DecodeFCPTemplate(const Fcp: string; var FileInfo: TFileInfo);
  var
    i,j,v             : integer;
    Tlv, SubTlv       : TTlvContainer;
    ChvMask, ChvIndex : integer;
    UsePin, UseUPin   : boolean;
    EnableUPin        : boolean;
    SecEnv            : integer;
    AccRuleRec        : integer;
  begin
    UsePin     := false;
    UseUPin    := false;
    EnableUPin := false;
    if Length(Fcp) > 1 then
      begin
      Tlv    := TTlvContainer.Create;
      SubTlv := TTlvContainer.Create;
      try
        SeparateTlvObjects(Fcp,Tlv);
        for i := Tlv.Count - 1 downto 0 do // the downto is absolutely neccessary
          case Tlv[i].Tag of
            TAGFileSize   : begin // File size ///////////////////////////////////////////////////////
                            if Length(Tlv[i].Value) = 2 then FileInfo.FileSize := BytesToCardinal(Tlv[i].Value)
                                                        else FileInfo.FileSize := 0;
                            end;
            TAGFileDesc   : begin // File Descriptor /////////////////////////////////////////////////
                            case OrdD(Tlv[i].Value,1) and $38 of
                              $00,
                              $08  : begin
                                     FileInfo.FileType   := ftElementary;
                                     case OrdD(Tlv[i].Value,1) and $07 of
                                       $01 : FileInfo.FileStruct := fsTransparent;
                                       $02 : FileInfo.FileStruct := fsLinearFixed;
                                       $06 : FileInfo.FileStruct := fsCyclic;
                                       else  FileInfo.FileStruct := fsUnknown;
                                       end;
                                     if FileInfo.FileStruct in [fsLinearFixed,fsCyclic] then
                                       begin
                                       FileInfo.RecordLen  := BytesToCardinal(Copy(Tlv[i].Value,3,2));
                                       FileInfo.NumRecords := OrdD(Tlv[i].Value,5);
                                       end
                                     else
                                       begin
                                       FileInfo.NumRecords := 0;
                                       end;
                                     end;
                              $38  : begin
                                     FileInfo.FileType   := ftDedicated;
                                     FileInfo.FileStruct := fsNone;
                                     FileInfo.NumRecords := 0;
                                     end;
                              //else
                              end;
                            end;
            TAGFileId     : begin // File Identifier /////////////////////////////////////////////////
                            if Length(Tlv[i].Value) = 2 then FileInfo.FileId := Tlv[i].Value
                                                        else FileInfo.FileId := '';
                            end;
            TAGAid        : begin  // DF name (AID) //////////////////////////////////////////////////
                            FileInfo.AppId := Tlv[i].Value;
                            end;
            TAGPropInfo   : begin // Proprietary information /////////////////////////////////////////
                            SeparateTlvObjects(Tlv[i].Value,SubTlv);
                            for j := 0 to SubTlv.Count - 1 do
                              case SubTlv[j].Tag of
                                $80 : begin
                                      v := BytesToCardinal(SubTlv[j].Value);
                                      case v and $0D of
                                        $00 : FileInfo.ClockStop := csNotAllowed;
                                        $04 : FileInfo.ClockStop := csOnlyHigh;
                                        $08 : FileInfo.ClockStop := csOnlyLow;
                                        $01 : FileInfo.ClockStop := csNoPref;
                                        $05 : FileInfo.ClockStop := csHighPref;
                                        $09 : FileInfo.ClockStop := csLowPref;
                                        else  FileInfo.ClockStop := csNotAllowed;
                                        end;
                                      if (v and $40) > 0 then FCardVoltage := vc18Volt
                                        else if (v and $20) > 0 then FCardVoltage := vc30Volt
                                          else if (v and $10) > 0 then FCardVoltage := vc50Volt
                                            else FCardVoltage := vcUnknown;
                                      end;
                                end;
                            end;
            TAGPinStatus  : begin //PIN Status Template DO
                            ChvMask  := 0;
                            ChvIndex := 0;
                            SeparateTlvObjects(Tlv[i].Value,SubTlv);
                            for j := 0 to SubTlv.Count - 1 do
                              case SubTlv[j].Tag of
                                $90 : begin
                                      ChvMask  := BytesToCardinal(SubTlv[j].Value);
                                      ChvIndex := 1 shl (Length(SubTlv[j].Value) * 8) + 7;
                                      end;
                                $83 : begin
                                      case OrdD(SubTlv[j].Value,1) of
                                        $01 : if ChvMask and ChvIndex > 0 then FileInfo.Chv1Enable := true
                                                                          else FileInfo.Chv1Enable := false;
                                        $11 : begin
                                              if UsePin then UseUPin := true
                                                        else UseUPin := false;
                                              if ChvMask and ChvIndex > 0 then EnableUPin := true
                                                                          else EnableUPin := false;
                                              end;
                                        end;
                                      ChvIndex := ChvIndex shr 1;
                                      Inc(FileInfo.NumCodes);
                                      UsePin := false;
                                      end;
                                $95 : begin
                                      if OrdD(SubTlv[j].Value,1) = $08 then UsePin := true
                                                                       else UsePin := false;
                                      end;
                                end;
                            end;
            TAGSecAttribB : begin
                            if FileInfo.Chv1Enable then SecEnv := 1
                              else if not EnableUPin then SecEnv := 0
                                else if UseUPin then SecEnv := 0
                                  else SecEnv := 1;
                            if Length(Tlv[i].Value) > 2 then
                              begin
                              FileInfo.AccessRules := Copy(Tlv[i].Value,1,2);
                              if Length(Tlv[i].Value) > 2 then
                                begin
                                AccRuleRec := OrdD(Tlv[i].Value,3);
                                end
                              else
                                begin
                                if OrdD(Tlv[i].Value,3) = SecEnv then AccRuleRec := OrdD(Tlv[i].Value,4)
                                  else if OrdD(Tlv[i].Value,5) = SecEnv then AccRuleRec := OrdD(Tlv[i].Value,6)
                                    else AccRuleRec := MAXINT;
                                end;
                              end
                            else
                              begin
                              FileInfo.AccessRules := '';
                              AccRuleRec := MAXINT;
                              end;
                            if FileInfo.AccessRules = FAccessRuleFile then
                              begin
                              if (AccRuleRec - 1) <= High(FAccessRules) then
                                begin
                                FileInfo.ReadCond :=  GetAccCondFromAccessRule(FAccessRules[AccRuleRec - 1].ReadAccess);
                                FileInfo.WriteCond := GetAccCondFromAccessRule(FAccessRules[AccRuleRec - 1].WriteAccess);
                                end
                              else
                                begin
                                FileInfo.ReadCond :=  acNever;
                                FileInfo.WriteCond := acNever;
                                end;
                              end;
                            end;

            {TAGTotFileSize: begin // Total File size ///////////////////////////////////////////////////////
                            end;}

            end;
        if FileInfo.FileType = ftDedicated then
          begin
          //RequestThePinRetryCounters
          end;
      finally
        Tlv.Free;
        SubTlv.Free;
      end;
      end;
    end;

{
2F00:
62
  22
    File Descriptor 82 05 02 21 00 23 01   Elementary, Linear fixed, 1 Record, $23Bytes
    FileID          83 02 2F 00
    Prop. Info      A5 03 80 01 35
    Life cycle      8A 01 05
    Security Attr.  8B 03 2F 06 05
    Filesize        80 02 00 23
    Tot. Filesize   81 02 00 42
    Short FID       88 00

3F00:
62
  2F
    File Descriptor 82 02 38 21
    FileID          83 02 3F 00
    AppID           84 04 72 6F 6F 74
    Prop. Info      A5 03 80 01 35
    Life cycle      8A 01 05
    Security Attr.  8B 03 2F 06 01
    C6 12
          90 01 70      PS_DO (Bitmap)
          83 01 01      PIN1
          83 01 0A      ADM1
          83 01 0B      ADM2
          83 01 0C      ADM3
          83 01 81      PIN2


The content of file 2F06 is as referenced in TS102221 - Annex E
  PIN1 is : Application PIN  - PIN        Appl.1 global (key reference '01');
  PIN2 is : Application PIN2 - Second PIN Appl.1 local  (key reference '81');
  UPIN is : Universal PIN    - UNIVERSAL PIN            (key reference '11');

  access condition ADM is                               (key reference '0A'..'0E' and
                                                                       '8A'..'8E')

ARR format:

 ARR         : (AM_TLV) (SC_TLV)
 SC_TLV      : 90    ALWAYS
               A4    (KEYREF_TLV) (USEQUAL_TLV)
               97    NEVER
 AM_TLV      : 80    READ > 01  UPDATE > 02
 KEYREF_TLV  : 83    key reference as stated above
 USEQUAL_TLV : 95    08 use PIN
                     00 requirement not used
                     (as in TS102221 - 9.5.2)
  }

begin
  ClearFileInfo(FileInfo);
  case FCardClass of
    ccGsmSim,
    ccJanus  : begin
               FileInfo.FileId := Copy(Status,5,2);
               case OrdD(Status,7) of
                 1 :  FileInfo.FileType := ftMaster;
                 2 :  FileInfo.FileType := ftDedicated;
                 4 :  FileInfo.FileType := ftElementary;
                 else FileInfo.FileType := ftUnknown;
                 end;
               case FileInfo.FileType of
                 ftElementary : DecodeFileStatusGsmEFile(Status,FileInfo);
                 ftMaster,
                 ftDedicated  : DecodeFileStatusGsmDFile(Status,FileInfo);
                 else if Assigned(FOnLogMessage) then FOnLogMessage(Self,'GetCardAttr','Error','Problem with FileStatus : ' + BinToHexExt(Status));
                 end;
               end;
    ccUicc,
    ccTSim,
    ccUSim   : begin
               if OrdD(Status,1) = TAGEtsiFcp then
                 begin
                 if OrdD(Status,2) < 127 then
                   begin
                   DecodeFCPTemplate(Copy(Status,3,OrdD(Status,2)),FileInfo);
                   end
                 else
                   begin //taken from ISO7816-4, not sure if this is applicable here
                   if OrdD(Status,2) = 129 then
                     DecodeFCPTemplate(Copy(Status,4,OrdD(Status,3)),FileInfo)
                     else if Assigned(FOnLogMessage) then FOnLogMessage(Self,'GetCardAttr','Error','Problem with FCP Template: ' + BinToHexExt(Status));
                   end;
                 end;
               end;
    end;
end;

function TPCSCConnector.DecodeResponseCode(const Code: word): string;
var
  sw1,sw2,temp : integer;
begin
sw1 := (Code and $FF00) shr 8;
sw2 := Code and $00FF;
Result := '';
case sw1 of
  $61 : Result := IntToStr(sw2) + ' bytes of data available via GET RESPONSE';
  $62 : case sw2 of
          $00 : Result := '(3G) No information given, state of non volatile memory unchanged';
          $81 : Result := '(3G) Part of returned data may be corrupted';
          $82 : Result := '(3G) End of file/record reached before reading Le bytes';
          $83 : Result := '(3G) Selected file invalidated';
          $85 : Result := '(ISO) Selected file in termination state';
          end;
  $63 : case sw2 of
          $C0..$CF : begin
                     temp := sw2 and $0F;
                     Result := '(3G) Command successful but after using an internal update retry routine ' + IntToStr(temp) + ' times'#13;
                     Result := Result + 'Verification failed, ' + IntToStr(temp) + ' retries remaining';
                     end;
          end;
  $64 : case sw2 of
          $00 : Result := '(3G) No information given, state of non-volatile memory unchanged';
          end;
  $65 : case sw2 of
          $00 : Result := '(3G) No information given, state of non-volatile memory changed';
          $81 : Result := '(3G) Memory problem';
          end;
  $67 : case sw2 of
          $00 : Result := '(GSM) Incorrect parameter P3'#13'(3G) Wrong length';
          else  Result := '(GSM) Incorrect parameter P3'#13'(3G) The interpretation of this status word is command dependent';
          end;
  $68 : case sw2 of
          $00 : Result := '(3G) No information given';
          $81 : Result := '(3G) Logical channel not supported';
          $82 : Result := '(3G) Secure messaging not supported';
          end;
  $69 : case sw2 of
          $00 : Result := '(3G) No information given';
          $81 : Result := '(3G) Command incompatible with file structure';
          $82 : Result := '(3G) Security status not satisfied';
          $83 : Result := '(3G) Authentication/PIN method blocked';
          $84 : Result := '(3G) Referenced data invalidated';
          $85 : Result := '(3G) Conditions of used not satisfied';
          $86 : Result := '(3G) Command not allowed (no EF selected)';
          end;
  $6A : case sw2 of
          $80 : Result := '(3G) Incorrect parameters in the data field';
          $81 : Result := '(3G) Function not supported';
          $82 : Result := '(3G) File not found';
          $83 : Result := '(3G) Record not found';
          $86 : Result := '(3G) Incorrect parameters P1-P2';
          $87 : Result := '(3G) Lc inconsistent with P1-P2';
          $88 : Result := '(3G) Referenced data not found';
          end;
  $6B : case sw2 of
          $00 : Result := 'Incorrect parameter P1 or P2';
          else  Result := '(GSM) Incorrect parameter P1 or P2';
          end;
  $6C : Result := 'Resend previous command with P2 set to '+ IntToStr(sw2);
  $6D : case sw2 of
          $00 : Result := 'Instruction code not supported or invalid';
          else  Result := '(GSM) Unknown instruction code given in the command';
          end;
  $6E : case sw2 of
          $00 : Result := 'Wrong instruction class given in the command';
          else  Result := '(GSM) Wrong instruction class given in the command';
          end;
  $6F : case sw2 of
          $00 : Result := 'Technical problem, no precise diagnosis'#13'Invalid input parameters to authentication computation';
          else  Result := '(GSM) Technical problem with no diagnostic given'#13 +
                          'The interpretation of this status word is command dependent';
          end;
  $90 : Result := 'Normal ending of the command';
  $91 : Result := 'Normal ending of the command, with extra information from the proactive UICC containing a command for the terminal';
  $92 : case sw2 of
          $00..$0F : Result := '(GSM) Command successful but after using an internal update retry routine ' + IntToStr(sw2) + ' times';
               $40 : Result := '(GSM) Memory problem';
          end;
  $93 : case sw2 of
          $00 : Result := 'SIM Application Toolkit is busy. Command cannot be executed at present, further normal commands are allowed';
          end;
  $94 : case sw2 of
          $00 : Result := '(GSM) No EF selected';
          $02 : Result := '(GSM) Out of range (invalid address)';
          $04 : Result := '(GSM) File ID or pattern not found';
          $08 : Result := '(GSM) File is inconsistent with the command';
          end;
  $98 : case sw2 of
          $02 : Result := '(GSM) No CHV initialized';
          $04 : Result := '(GSM) Access condition not fulfilled or'#13 +
                          'Unsuccessful UNBLOCK/CHV verification, at least one attempt left or'#13 +
                          'Authentication failed';
          $08 : Result := '(GSM) In contradiction with CHV status';
          $10 : Result := '(GSM) In contradiction with invalidation status';
          $34 : result := '(GSM) Update SSD order sequence not respected';
          $40 : Result := '(GSM) Unsuccessful UNBLOCK/CHV verification, no attempt left or'#13 +
                          '(GSM) UNBLOCK/CHV blocked';
          $50 : Result := 'Increase cannot be performed, Max value reached';
          $60 : Result := '(TETRA) Manipulation flag set';
          $62 : Result := '(3G) Authentication error, application specific';
          $70 : Result := '(TETRA) SwMI authentication unsuccessful';
          end;
  $9E : Result := '(GSM) ' + IntToStr(sw2) + ' bytes of response data given in case of a SIM data download error';
  $9F : Result := IntToStr(sw2) + ' bytes of response data available';
  end;

if Result = '' then Result := 'Unknown result code ' + IntToHex(Code,4);
end;

function TPCSCConnector.DecodeSCardReturnCode(const Code: cardinal): string;
begin
  case Code of
    SCARD_S_SUCCESS                 : Result := 'SCARD_S_SUCCESS : Everything is OK';
    SCARD_E_BAD_SEEK                : Result := 'SCARD_E_BAD_SEEK : There was an error trying to set the smart card file object pointer.';
    SCARD_E_CANCELLED               : Result := 'SCARD_E_CANCELLED : The action was cancelled by an SCardCancel request.';
    SCARD_E_CANT_DISPOSE            : Result := 'SCARD_E_CANT_DISPOSE : The system could not dispose of the media in the requested manner.';
    SCARD_E_CARD_UNSUPPORTED        : Result := 'SCARD_E_CARD_UNSUPPORTED : The smart card does not meet minimal requirements for support.';
    SCARD_E_CERTIFICATE_UNAVAILABLE : Result := 'SCARD_E_CERTIFICATE_UNAVAILABLE : The requested certificate could not be obtained.';
    SCARD_E_COMM_DATA_LOST          : Result := 'SCARD_E_COMM_DATA_LOST : A communications error with the smart card has been detected. Retry the operation.';
    SCARD_E_DIR_NOT_FOUND           : Result := 'SCARD_E_DIR_NOT_FOUND : The identified directory does not exist in the smart card.';
    SCARD_E_DUPLICATE_READER        : Result := 'SCARD_E_DUPLICATE_READER : The reader driver did not produce a unique reader name.';
    SCARD_E_FILE_NOT_FOUND          : Result := 'SCARD_E_FILE_NOT_FOUND : The identified file does not exist in the smart card.';
    SCARD_E_ICC_CREATEORDER         : Result := 'SCARD_E_ICC_CREATEORDER : The requested order of object creation is not supported.';
    SCARD_E_ICC_INSTALLATION        : Result := 'SCARD_E_ICC_INSTALLATION : No Primary Provider can be found for the smart card.';
    SCARD_E_INSUFFICIENT_BUFFER     : Result := 'SCARD_E_INSUFFICIENT_BUFFER : The data buffer to receive returned data is too small for the returned data.';
    SCARD_E_INVALID_ATR             : Result := 'SCARD_E_INVALID_ATR : An ATR obtained from the registry is not a valid ATR string.';
    SCARD_E_INVALID_CHV             : Result := 'SCARD_E_INVALID_CHV : The supplied PIN is incorrect.';
    SCARD_E_INVALID_HANDLE          : Result := 'SCARD_E_INVALID_HANDLE : The supplied handle was invalid.';
    SCARD_E_INVALID_PARAMETER       : Result := 'SCARD_E_INVALID_PARAMETER : One or more of the supplied parameters could not be properly interpreted.';
    SCARD_E_INVALID_TARGET          : Result := 'SCARD_E_INVALID_TARGET : Registry startup information is missing or invalid.';
    SCARD_E_INVALID_VALUE           : Result := 'SCARD_E_INVALID_VALUE : One or more of the supplied parameters values could not be properly interpreted.';
    SCARD_E_NO_ACCESS               : Result := 'SCARD_E_NO_ACCESS : Access is denied to this file.';
    SCARD_E_NO_DIR                  : Result := 'SCARD_E_NO_DIR : The supplied path does not represent a smart card directory.';
    SCARD_E_NO_FILE                 : Result := 'SCARD_E_NO_FILE : The supplied path does not represent a smart card file.';
    SCARD_E_NO_KEY_CONTAINER        : Result := 'SCARD_E_NO_KEY_CONTAINER : The requested key container does not exist on the smart card.';
    SCARD_E_NO_MEMORY               : Result := 'SCARD_E_NO_MEMORY : Not enough memory available to complete this command.';
    SCARD_E_NO_READERS_AVAILABLE    : Result := 'SCARD_E_NO_READERS_AVAILABLE : Cannot find a smart card reader.';
    SCARD_E_NO_SERVICE              : Result := 'SCARD_E_NO_SERVICE : The Smart card resource manager is not running.';
    SCARD_E_NO_SMARTCARD            : Result := 'SCARD_E_NO_SMARTCARD : The operation requires a Smart Card, but no Smart Card is currently in the device.';
    SCARD_E_NO_SUCH_CERTIFICATE     : Result := 'SCARD_E_NO_SUCH_CERTIFICATE : The requested certificate does not exist.';
    SCARD_E_NOT_READY               : Result := 'SCARD_E_NOT_READY : The reader or smart card is not ready to accept commands.';
    SCARD_E_NOT_TRANSACTED          : Result := 'SCARD_E_NOT_TRANSACTED : An attempt was made to end a non-existent transaction.';
    SCARD_E_PCI_TOO_SMALL           : Result := 'SCARD_E_PCI_TOO_SMALL : The PCI Receive buffer was too small.';
    SCARD_E_PROTO_MISMATCH          : Result := 'SCARD_E_PROTO_MISMATCH : The requested protocols are incompatible with the protocol currently in use with the smart card.';
    SCARD_E_READER_UNAVAILABLE      : Result := 'SCARD_E_READER_UNAVAILABLE : The specified reader is not currently available for use.';
    SCARD_E_READER_UNSUPPORTED      : Result := 'SCARD_E_READER_UNSUPPORTED : The reader driver does not meet minimal requirements for support.';
    SCARD_E_SERVICE_STOPPED         : Result := 'SCARD_E_SERVICE_STOPPED : The Smart card resource manager has shut down.';
    SCARD_E_SHARING_VIOLATION       : Result := 'SCARD_E_SHARING_VIOLATION : The smart card cannot be accessed because of other connections outstanding.';
    SCARD_E_SYSTEM_CANCELLED        : Result := 'SCARD_E_SYSTEM_CANCELLED : The action was cancelled by the system, presumably to log off or shut down.';
    SCARD_E_TIMEOUT                 : Result := 'SCARD_E_TIMEOUT : The user-specified timeout value has expired.';
    SCARD_E_UNEXPECTED              : Result := 'SCARD_E_UNEXPECTED : An unexpected card error has occurred.';
    SCARD_E_UNKNOWN_CARD            : Result := 'SCARD_E_UNKNOWN_CARD : The specified smart card name is not recognized.';
    SCARD_E_UNKNOWN_READER          : Result := 'SCARD_E_UNKNOWN_READER : The specified reader name is not recognized.';
    SCARD_E_UNKNOWN_RES_MNG         : Result := 'SCARD_E_UNKNOWN_RES_MNG : An unrecognized error code was returned from a layered component.';
    SCARD_E_UNSUPPORTED_FEATURE     : Result := 'SCARD_E_UNSUPPORTED_FEATURE : This smart card does not support the requested feature.';
    SCARD_E_WRITE_TOO_MANY          : Result := 'SCARD_E_WRITE_TOO_MANY : The smartcard does not have enough memory to store the information.';
    SCARD_F_COMM_ERROR              : Result := 'SCARD_F_COMM_ERROR : An internal communications error has been detected.';
    SCARD_F_INTERNAL_ERROR          : Result := 'SCARD_F_INTERNAL_ERROR : An internal consistency check failed.';
    SCARD_F_UNKNOWN_ERROR           : Result := 'SCARD_F_UNKNOWN_ERROR : An internal error has been detected, but the source is unknown.';
    SCARD_F_WAITED_TOO_LONG         : Result := 'SCARD_F_WAITED_TOO_LONG : An internal consistency timer has expired.';
    SCARD_P_SHUTDOWN                : Result := 'SCARD_P_SHUTDOWN : The operation has been aborted to allow the server application to exit.';
    SCARD_W_CANCELLED_BY_USER       : Result := 'SCARD_W_CANCELLED_BY_USER : The action was cancelled by the user.';
    SCARD_W_CARD_NOT_AUTHENTICATED  : Result := 'SCARD_W_CARD_NOT_AUTHENTICATED : No PIN was presented to the smart card.';
    SCARD_W_CHV_BLOCKED             : Result := 'SCARD_W_CHV_BLOCKED : The card cannot be accessed because the maximum number of PIN entry attempts has been reached.';
    SCARD_W_EOF                     : Result := 'SCARD_W_EOF : The end of the smart card file has been reached.';
    SCARD_W_REMOVED_CARD            : Result := 'SCARD_W_REMOVED_CARD : The smart card has been removed, so that further communication is not possible.';
    SCARD_W_RESET_CARD              : Result := 'SCARD_W_RESET_CARD : The smart card has been reset, so any shared state information is invalid.';
    SCARD_W_SECURITY_VIOLATION      : Result := 'SCARD_W_SECURITY_VIOLATION : Access was denied because of a security violation.';
    SCARD_W_UNPOWERED_CARD          : Result := 'SCARD_W_UNPOWERED_CARD : Power has been removed from the smart card, so that further communication is not possible.';
    SCARD_W_UNRESPONSIVE_CARD       : Result := 'SCARD_W_UNRESPONSIVE_CARD : The smart card is not responding to a reset.';
    SCARD_W_UNSUPPORTED_CARD        : Result := 'SCARD_W_UNSUPPORTED_CARD : The reader cannot communicate with the smart card, due to ATR configuration conflicts.';
    SCARD_W_WRONG_CHV               : Result := 'SCARD_W_WRONG_CHV : The card cannot be accessed because the wrong PIN was presented.';
    SCARD_E_PCSC_NOTLOADED          : Result := 'SCARD_E_PCSC_NOTLOADED : The PC/SC software layer is not available';
    SCARD_E_INVALID_CONTEXT         : Result := 'SCARD_E_INVALID_CONTEXT : The specified card manager context is not valid';
    $00000032                       : Result := 'ERROR_NOT_SUPPORTED : The specified parameter is not supported by the device';
    $00000006                       : Result := 'USER_E_INVALIDHANDLE : The specified handle is invalid';
    $00000057                       : Result := 'USER_E_TIMEOUT : The session with the card got a timeout, the command was not completed within a given timeframe.';
    else                              Result := IntToHex(Code,8) + ' : Unknown result code';
    end;
end;

end.

